!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>                                                                     >
!>                                                                     >
!>         Tp_Main.f95: Code unit including the main program           >
!>                      and all related routines                       >
!>                                                                     >
!>                                                                     >
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!
!
!
      MODULE MeshMaker_Data
! 
         SAVE
! 
! ----------
! ...... Double precision variables
! ----------
! 
         REAL(KIND = 8) :: X_ref, Y_ref, Z_ref, factor
! 
! ----------
! ...... Integer parameters
! ----------
! 
         INTEGER, PARAMETER :: MESH_Unit = 12   
         INTEGER, PARAMETER :: MINC_Unit = 14   
! 
! ----------
! ...... Common integer variables      
! ----------
! 
         INTEGER :: ElemNameLength
! 
         INTEGER :: Max_NumElem
! 
         INTEGER :: Longest
! 
! ----------
! ...... Character arrays
! ----------
! 
         CHARACTER(LEN = 3) :: FormatType, LengthUnits 
! 
         CHARACTER(LEN = 8), ALLOCATABLE, DIMENSION(:) :: name
! 
! ----------
! ...... Logical variables
! ----------
! 
         LOGICAL :: media_by_number = .FALSE.
! 
! ----------
! ...... Logical arrays
! ----------
! 
         LOGICAL, ALLOCATABLE, DIMENSION(:) :: fractured
! 
! 
      END MODULE MeshMaker_Data
!
!
!
!***********************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***********************************************************************
!
!
!
      PROGRAM T_MeshMaker
! 
! ...... Modules to be used 
! 
         USE MeshMaker_Data
!
!***********************************************************************
!***********************************************************************
!***********************************************************************
!*                                                                     *
!*      TOUGH+ Main Program: Calls several lower-level routines        *     
!*                    which execute computations                       *     
!*                                                                     *
!*                   Version 1.0 - April 18, 2004                      *     
!*                                                                     *
!***********************************************************************
!***********************************************************************
!***********************************************************************
!
      IMPLICIT NONE
! 
! -------
! ... Integer Variables    
! -------
! 
      INTEGER :: ierG
! 
! -------
! ... Character Variables    
! -------
! 
      CHARACTER(LEN = 120) :: TITLE
! 
! -------
! ... Logical Variables    
! -------
! 
      LOGICAL :: exists = .FALSE., Flag_MINC
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>  Main body of TOUGH+
! 
      WRITE(*,6000) 
! 
! -------
! ... Inquire whether file MESH exists; open accordingly
! -------
! 
      INQUIRE(FILE='MESH', EXIST = exists)
! 
      IF_MESH: IF(exists) THEN
                  WRITE(*,6004)
                  OPEN(MESH_Unit, FILE = 'MESH', STATUS = 'OLD')
               ELSE
                  WRITE(*,6005)
                  OPEN(MESH_Unit, FILE = 'MESH', STATUS = 'NEW')
               END IF IF_MESH
! 
      REWIND (UNIT = MESH_Unit)
! 
! -------
! ... Reading the heading/title of the input data file  
! -------
! 
      READ (*,6002) TITLE
      WRITE(*,6003) TITLE
      
      LengthUnits = ' '
!
! -------
! ... Read the number of elements, connections and 
! ... number of characters in the element name
! -------
!
      READ(*,*) Max_NumElem, Longest, ElemNameLength, FormatType, LengthUnits, media_by_number
!
      IF(ElemNameLength /= 8) ElemNameLength = 5
!
      SELECT CASE (LengthUnits)
      CASE('m', 'M')
         factor = 1.0d0
      CASE('ft', 'FT', 'Ft')
         factor = 3.048d-1
      CASE('in', 'IN', 'In')
         factor = 2.54d-2
      CASE('km', 'KM', 'Km')
         factor = 1.0d3
      CASE('cm', 'CM', 'Cm')
         factor = 1.0d-2
      CASE('mi', 'MI', 'Mi')
         factor = 1.609344d3
      CASE DEFAULT
         WRITE(UNIT = *, FMT = 6101) 'length', LengthUnits
         STOP
      END SELECT
!
      ALLOCATE(name(Max_NumElem), fractured(0:Max_NumElem), STAT = ierG)
!
! -------
! ... End of inputs
! -------
!
      WRITE(*,6006) 
! 
! 
!***********************************************************************
!*                                                                     *
!*                  CALL THE MESH GENERATION PACKAGE                   *
!*                                                                     *
!***********************************************************************
! 
! 
      CALL Mesh_Maker(Flag_MINC)
!
!
!***********************************************************************
!*                                                                     *
!*                        F  O  R  M  A  T  S                          *
!*                                                                     *
!***********************************************************************
!
!
!6000 FORMAT('MeshMaker         1.0   10 October   2006',//)
 6000 FORMAT('MeshMaker         1.5   15 August    2014',//)
!
 6002 FORMAT(A120)
 6003 FORMAT(/1X,131('='),//,5X,'PROBLEM TITLE:  ',A120,//)


!
 6004 FORMAT(' FILE "MESH" EXISTS ==> OPEN AS AN OLD FILE')
 6005 FORMAT(' FILE "MESH" DOES NOT EXIST ==> OPEN AS A NEW FILE')
!
 6006 FORMAT(//,' END OF MeshMaker INPUT JOB',/)
!
 6101 FORMAT(//,20('ERROR-'),//,   &
     &       T10,'>>>  The units of <',A,'> = ',A,' are not an available option',//,  &
     &       T10,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',      &
     &       //,20('ERROR-'))
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>  End of T_MeshMaker
!
!
      STOP
!
!
!
      END PROGRAM T_MeshMaker
!
!
!
!***********************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***********************************************************************




!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>                                                                     >
!>                                                                     >
!>            Module including grid generation parameters              >
!>                                                                     >
!>                                                                     >
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!
!
!
!
      MODULE Grid_Gener_parameters
! 
         SAVE
! 
! ----------
! ...... Double precision parameters
! ----------
! 
         REAL(KIND = 8), PARAMETER :: pi  = 3.14159265358979324D0  
! 
! ----------
! ...... Character parameter arrays     
! ----------
! 
         CHARACTER(LEN = 1), PARAMETER, &
     &                       DIMENSION(36) :: NA = (/ '0','1','2','3','4','5','6','7','8','9',   &
     &                                                'A','B','C','D','E','F','G','H','I','J',   &
     &                                                'K','L','M','N','O','P','Q','R','S','T',   &
     &                                                'U','V','W','X','Y','Z'  /)
! 
         CHARACTER(LEN = 5), PARAMETER, &
     &                       DIMENSION(6) :: TYPOS = (/ 'ONE-D','TWO-D','THRED','STANA','STANB','STANT' /)
! 
! ----------
! ...... Integer variables
! ----------
! 
         INTEGER :: NSKIP,NELEMT,NLAY
! 
! ----------
! ...... CHARACTER variables
! ----------
! 
         CHARACTER(LEN = 4) :: WHE_x
         CHARACTER(LEN = 5) :: dominant_medium
! 
! ----------
! ...... Logical variables
! ----------
! 
         LOGICAL :: Flag_HetRegions,Flag_Boundaries,Flag_ExclZones
! 
! 
! 
      END MODULE Grid_Gener_parameters
!
!
!
!***********************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***********************************************************************
!
!
!
      MODULE Het_Region_Definition
! 
! 
         SAVE
! 
! ----------
! ...... Derived-type variables     
! ----------
! 
         TYPE Het_Region
!                                      
            CHARACTER(LEN = 5)           :: name                   ! Name of the heterogeneous region 
!                                      
            CHARACTER(LEN = 11)          :: coordinates            ! Name of the heterogeneous region 
!                                      
            CHARACTER(LEN = 2 )          :: units                  ! Name of the heterogeneous region 
! 
            REAL(KIND = 8), DIMENSION(3) :: L_min,L_max            ! Min and max extent of the region in the principal directions 
            REAL(KIND = 8)               :: x0,y0,z0,R_min,R_max   ! Circle/sphere: x,y,z center cordinates, ... 
!                                                                  ! ... and min and max radius (for mixed coordinates)
         END TYPE Het_Region
! 
! ----------
! ...... Derived-type arrays 
! ----------
! 
         TYPE(Het_Region), ALLOCATABLE, DIMENSION(:) :: H_Region
! 
! ----------
! ...... Integer variables
! ----------
! 
         INTEGER :: Num_HetRegions
! 
! >>>>> 
! >>>>> 
! >>>>> 
! >>>>> 
! >>>>> 
! 
         CONTAINS
!
!
!
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
!
!
!
         CHARACTER(LEN = 5) FUNCTION Media_Region(Flag_HetRegions,coordinates,dominant_medium,X,Y,Z,media_by_number)
!
!***********************************************************************
!***********************************************************************
!*                                                                     *
!*             Routine determining the media region where              *
!*                     a particular cell belongs                       *
!*                                                                     *
!***********************************************************************
!***********************************************************************
!
            IMPLICIT NONE
! 
! -------------
! ......... Double precision variables
! -------------
! 
            REAL(KIND = 8) :: X,Y,Z, R_test 
! 
! -------------
! ......... Integer variables
! -------------
! 
            INTEGER :: n, N_0, N_1, N_2, N_3, N_4, N_5
! 
! -------------
! ......... CHARACTER variables
! -------------
! 
            CHARACTER(LEN =  5) :: dominant_medium
            CHARACTER(LEN = 11) :: coordinates
! 
! -------------
! ......... Logical variables
! -------------
! 
            LOGICAL, INTENT(IN) :: media_by_number
! 
            LOGICAL :: Flag_HetRegions
! 
            LOGICAL :: First_call = .TRUE.
! 
! -------------
! ......... Saving variables
! -------------
! 
            SAVE First_call
!
!
! =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   Main body of Media_Region
!
!
            IF(First_call) THEN
               First_call = .FALSE.
               WRITE(*,6000)
            END IF
! 
! -------------
! ......... Initialization
! -------------
! 
            Media_Region = dominant_medium  ! Initial assignment to the first domain (considered dominant by default)
!
            IF(media_by_number) THEN
               IF( dominant_medium(1:1) == 'F' .OR. dominant_medium(1:1) == 'f' ) THEN
                  Media_Region = ' 0001'
               ELSE
                  Media_Region = '-0001'
               END IF
            END IF
!
!
!***********************************************************************
!*                                                                     *
!*          For cartesian coordinate systems                           *
!*                                                                     *
!***********************************************************************
!
!
            IF_Assign: IF(coordinates(1:4) == 'CART' .OR. coordinates(1:4) == 'cart' .OR. coordinates(1:4) == 'Cart') THEN
! 
               DO_Cartesian: DO n=2,Num_HetRegions
! 
! ............... Cartesian grid system, Cartesian-coordinate-defined media regions 
! 
                  IF_Range1: IF( (H_Region(n)%coordinates(1:4) == 'CART' .OR. H_Region(n)%coordinates(1:4) == 'cart' .OR. &
     &                            H_Region(n)%coordinates(1:4) == 'Cart') .AND.                     & 
     &                            X >= H_Region(n)%L_min(1) .AND. X <= H_Region(n)%L_max(1) .AND.   &          
     &                            Y >= H_Region(n)%L_min(2) .AND. Y <= H_Region(n)%L_max(2) .AND.   &
     &                            Z >= H_Region(n)%L_min(3) .AND. Z <= H_Region(n)%L_max(3)       ) & 
     &            THEN
! 
                     Media_Region = H_Region(n)%name
! 
! .................. Listing by material number 
! 
                     IF(media_by_number) THEN
! 
                        N_0 = n / 1000
                        N_1 = MOD(n,1000)
                        N_2 = N_1 /100
                        N_3 = MOD(N_1,100)
                        N_4 = N_3 / 10
                        N_5 = MOD(N_3,10)
! 
                        IF( Media_Region(1:1) == 'F' .OR. Media_Region(1:1) == 'f' ) THEN
                           Media_Region = ' '//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        ELSE
                           Media_Region = '-'//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        END IF
! 
                     END IF
! 
                  END IF IF_Range1
! 
! ............... Cartesian grid system, mixed-coordinate-defined media regions 
! 
                  IF_Range2: IF( (H_Region(n)%coordinates(1:4) == 'MIXE' .OR.    &
     &                            H_Region(n)%coordinates(1:4) == 'mixe' .OR.    &
     &                            H_Region(n)%coordinates(1:4) == 'Mixe') .AND.  & 
     &                            X >= H_Region(n)%L_min(1) .AND. X <= H_Region(n)%L_max(1) .AND.  &          
     &                            Y >= H_Region(n)%L_min(2) .AND. Y <= H_Region(n)%L_max(2) .AND.  &
     &                            Z >= H_Region(n)%L_min(3) .AND. Z <= H_Region(n)%L_max(3) .AND.  &
     &                            SQRT(   (X-H_Region(n)%x0)**2                                    & 
     &                                  + (Y-H_Region(n)%y0)**2                                    &   
     &                                  + (Z-H_Region(n)%z0)**2  ) >= H_Region(n)%R_min .AND.      &
     &                                SQRT(   (X-H_Region(n)%x0)**2                                & 
     &                                      + (Y-H_Region(n)%y0)**2                                &   
     &                                      + (Z-H_Region(n)%z0)**2  ) <= H_Region(n)%R_max      ) &
     &            THEN
! 
                     Media_Region = H_Region(n)%name
! 
! .................. Listing by material number 
! 
                     IF(media_by_number) THEN
! 
                        N_0 = n / 1000
                        N_1 = MOD(n,1000)
                        N_2 = N_1 /100
                        N_3 = MOD(N_1,100)
                        N_4 = N_3 / 10
                        N_5 = MOD(N_3,10)
! 
                        IF( Media_Region(1:1) == 'F' .OR. Media_Region(1:1) == 'f' ) THEN
                           Media_Region = ' '//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        ELSE
                           Media_Region = '-'//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        END IF
! 
                     END IF
! 
                  END IF IF_Range2
! 
! ............... Cartesian grid system, cylindrical-coordinate-defined media regions : ERROR !!! 
! 
                  IF_Range3: IF( (H_Region(n)%coordinates(1:4) == 'CYLI' .OR. H_Region(n)%coordinates(1:4) == 'cyli' .OR. &
     &                            H_Region(n)%coordinates(1:4) == 'Cyli') )  &
     &            THEN
! 
! .................. Check the various options 
! 
                     IF_CylOptions: IF( (H_Region(n)%z0 > 1.0d10) .AND. (H_Region(n)%x0 < 1.0d10) .AND.  &
     &                                  (H_Region(n)%y0 < 1.0d10)) &
     &               THEN
! 
! ..................... Cylinder axis along z 
! 
                        R_test = SQRT( (X-H_Region(n)%x0)**2 + (Y-H_Region(n)%y0)**2 )
                        IF((R_test >= H_Region(n)%R_min) .AND. (R_test <= H_Region(n)%R_max)) Media_Region = H_Region(n)%name
! 
                     ELSE IF((H_Region(n)%y0 > 1.0d10) .AND. (H_Region(n)%x0 < 1.0d10) .AND. (H_Region(n)%z0 < 1.0d10)) THEN
! 
! ..................... Cylinder axis along y 
! 
                        R_test = SQRT( (X-H_Region(n)%x0)**2 + (Y-H_Region(n)%z0)**2 )
                        IF((R_test >= H_Region(n)%R_min) .AND. (R_test <= H_Region(n)%R_max)) Media_Region = H_Region(n)%name
! 
                     ELSE IF((H_Region(n)%x0 > 1.0d10) .AND. (H_Region(n)%z0 < 1.0d10) .AND. (H_Region(n)%y0 < 1.0d10)) THEN
! 
! ..................... Cylinder axis along x 
! 
                        R_test = SQRT( (X-H_Region(n)%y0)**2 + (Y-H_Region(n)%z0)**2 )
                        IF((R_test >= H_Region(n)%R_min) .AND. (R_test <= H_Region(n)%R_max)) Media_Region = H_Region(n)%name
! 
                     ELSE 
! 
! ..................... Error  
! 
                        WRITE(*,6108) n
                        STOP
! 
                     END IF IF_CylOptions
! 
                  END IF IF_Range3
! 
               END DO DO_Cartesian
! 
               RETURN
!
!***********************************************************************
!*                                                                     *
!*          For cylindrical coordinate systems                         *
!*                                                                     *
!***********************************************************************
!
            ELSE IF(coordinates(1:4) == 'CYLI' .OR. coordinates(1:4) == 'cyli' .OR. coordinates(1:4) == 'Cyli') THEN
! 
               DO_Cylindrical: DO n=2,Num_HetRegions
! 
! ............... Cylindrical grid system, cylindrical-coordinate-defined media regions 
! 
                  IF_Range4: IF( (H_Region(n)%coordinates(1:4) == 'CYLI' .OR.    &
     &                            H_Region(n)%coordinates(1:4) == 'cyli' .OR.    &
     &                            H_Region(n)%coordinates(1:4) == 'Cyli') .AND.  &  
     &                            X >= H_Region(n)%L_min(1) .AND. X <= H_Region(n)%L_max(1) .AND.   &          
     &                            Z >= H_Region(n)%L_min(3) .AND. Z <= H_Region(n)%L_max(3)       ) & 
     &            THEN
! 
                     Media_Region = H_Region(n)%name
! 
! .................. Listing by material number 
! 
                     IF(media_by_number) THEN
! 
                        N_0 = n / 1000
                        N_1 = MOD(n,1000)
                        N_2 = N_1 /100
                        N_3 = MOD(N_1,100)
                        N_4 = N_3 / 10
                        N_5 = MOD(N_3,10)
! 
                        IF( Media_Region(1:1) == 'F' .OR. Media_Region(1:1) == 'f' ) THEN
                           Media_Region = ' '//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        ELSE
                           Media_Region = '-'//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        END IF
! 
                     END IF
! 
                  END IF IF_Range4
! 
! ............... Cylindrical grid system, other than cylindrical-coordinate-defined media regions : ERROR !!! 
! 
                  IF_Range5: IF( (H_Region(n)%coordinates(1:4) == 'CART' .OR. H_Region(n)%coordinates(1:4) == 'cart' .OR. &
     &                            H_Region(n)%coordinates(1:4) == 'Cart') .OR.                                            &
     &                           (H_Region(n)%coordinates(1:4) == 'MIXE' .OR. H_Region(n)%coordinates(1:4) == 'mixe' .OR. &
     &                            H_Region(n)%coordinates(1:4) == 'Mixe') )  &
     &            THEN
                     WRITE(*,6102) n,H_Region(n)%coordinates
                     STOP
                  END IF IF_Range5
! 
               END DO DO_Cylindrical
! 
               RETURN
! 
! 
! 
            END IF IF_Assign
!
!
!***********************************************************************
!*                                                                     *
!*                        F  O  R  M  A  T  S                          *
!*                                                                     *
!***********************************************************************
!
!
!6000 FORMAT(/,'Media_Region      1.0   30 August    2005',6X,'Assigning cells to regions of heterogeneous media')
 6000 FORMAT(/,'Media_Region      1.0    3 August    2014',6X,'Assigning cells to regions of heterogeneous media')
!
 6100 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The coordinate system of the grid is Cartesian, ', &
     &          'but the coordinates defining the heterogeneous medium region #',I3,', are ',A11,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
 6102 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The coordinate system of the grid is cylindrical, ', &
     &          'but the coordinates defining the heterogeneous medium region #',I3,', are ',A11,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
 6108 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The coordinate system of the grid is Cartesian, ', &
     &          'but the axis of the cylindrical heterogeneous medium region #',I3,', is undefined',//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
!
! =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of Media_Region
!
!
            RETURN
!
         END FUNCTION Media_Region
!
!
!
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
! 
! 
! 
      END MODULE Het_Region_Definition
!
!
!
!***********************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***********************************************************************
!
!
!
      MODULE Boundary_Definition
! 
         USE Het_Region_Definition
! 
         SAVE
! 
! ----------
! ...... Derived-type variables     
! ----------
! 
         TYPE Boundary
!                                      
            CHARACTER(LEN = 1)  :: id             ! Identifier of the boundary 
!                                      
            TYPE(Het_Region)    :: region         ! Name of the heterogeneous region 
!                                      
         END TYPE Boundary
! 
! ----------
! ...... Derived-type arrays 
! ----------
! 
         TYPE(Boundary), ALLOCATABLE, DIMENSION(:) :: bound
! 
! ----------
! ...... Character arrays
! ----------
! 
         CHARACTER(LEN = 1), ALLOCATABLE, DIMENSION(:) :: bndry_id
! 
! ----------
! ...... Integer variables
! ----------
! 
         INTEGER :: Num_Boundaries
! 
! >>>>> 
! >>>>> 
! >>>>> 
! >>>>> 
! >>>>> 
! 
         CONTAINS
!
!
!
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
!
!
!
         SUBROUTINE Boundary_Status(coordinates,X,Y,Z,bound_id,medium, media_by_number)
!
!***********************************************************************
!***********************************************************************
!*                                                                     *
!*      Routine determining whether a cell belongs to a boundary       *
!*                                                                     *
!***********************************************************************
!***********************************************************************
!
            IMPLICIT NONE
! 
! -------------
! ......... Double precision variables
! -------------
! 
            REAL(KIND = 8) :: X,Y,Z 
! 
! -------------
! ......... Integer variables
! -------------
! 
            INTEGER :: n, n_t, N_0, N_1, N_2, N_3, N_4, N_5
! 
! -------------
! ......... CHARACTER variables
! -------------
! 
            CHARACTER(LEN =  1) :: bound_id
            CHARACTER(LEN =  5) :: medium
            CHARACTER(LEN = 11) :: coordinates
! 
! -------------
! ......... Logical variables
! -------------
! 
            LOGICAL, INTENT(IN) :: media_by_number
! 
            LOGICAL :: First_call = .TRUE.
! 
! -------------
! ......... Saving variables
! -------------
! 
            SAVE First_call
!
!
! =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>  Main body of Boundary_Status
!
!
            IF(First_call) THEN
               First_call = .FALSE.
               WRITE(*,6000)
            END IF
! 
! -------------
! ......... Initialization
! -------------
! 
            bound_id  = ' '      ! Initial assignment to active status
!
!
!***********************************************************************
!*                                                                     *
!*          For cartesian coordinate systems                           *
!*                                                                     *
!***********************************************************************
!
!
            IF_Assign: IF(coordinates(1:4) == 'CART' .OR. coordinates(1:4) == 'cart' .OR. coordinates(1:4) == 'Cart') THEN
! 
               DO_Cartesian: DO n=1,Num_Boundaries
! 
! ............... Cartesian grid system, Cartesian-coordinate-defined boundaries 
! 
                  IF_Range1: IF( (bound(n)%region%coordinates(1:4) == 'CART' .OR.     &
     &                            bound(n)%region%coordinates(1:4) == 'cart' .OR.     &
     &                            bound(n)%region%coordinates(1:4) == 'Cart') .AND.   & 
     &                            X >= bound(n)%region%L_min(1) .AND. X <= bound(n)%region%L_max(1) .AND.   &          
     &                            Y >= bound(n)%region%L_min(2) .AND. Y <= bound(n)%region%L_max(2) .AND.   &
     &                            Z >= bound(n)%region%L_min(3) .AND. Z <= bound(n)%region%L_max(3)       ) & 
     &            THEN
! 
                     medium   = bound(n)%region%name
                     bound_id = bound(n)%id
! 
! .................. Listing by material number 
! 
                     IF(media_by_number) THEN
! 
                        n_t = Num_HetRegions + n
                        N_0 = n_t / 1000
                        N_1 = MOD(n_t,1000)
                        N_2 = N_1 /100
                        N_3 = MOD(N_1,100)
                        N_4 = N_3 / 10
                        N_5 = MOD(N_3,10)
! 
                        IF( medium(1:1) == 'F' .OR. medium(1:1) == 'f' ) THEN
                           medium = ' '//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        ELSE
                           medium = '-'//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        END IF
! 
                     END IF
! 
                  END IF IF_Range1
! 
! ............... Cartesian grid system, mixed-coordinate-defined media boundaries 
! 
                  IF_Range2: IF( (bound(n)%region%coordinates(1:4) == 'MIXE' .OR.    &
     &                            bound(n)%region%coordinates(1:4) == 'mixe' .OR.    &
     &                            bound(n)%region%coordinates(1:4) == 'Mixe') .AND.  & 
     &                            X >= bound(n)%region%L_min(1) .AND. X <= bound(n)%region%L_max(1) .AND.  &          
     &                            Y >= bound(n)%region%L_min(2) .AND. Y <= bound(n)%region%L_max(2) .AND.  &
     &                            Z >= bound(n)%region%L_min(3) .AND. Z <= bound(n)%region%L_max(3) .AND.  &
     &                            SQRT(   (X-bound(n)%region%x0)**2                                        & 
     &                                  + (Y-bound(n)%region%y0)**2                                        &   
     &                                  + (Z-bound(n)%region%z0)**2  ) >= bound(n)%region%R_min .AND.      &
     &                            SQRT(   (X-bound(n)%region%x0)**2                                    & 
     &                                  + (Y-bound(n)%region%y0)**2                                    &   
     &                                  + (Z-bound(n)%region%z0)**2  ) <= bound(n)%region%R_max      ) &
     &            THEN
                     medium   = bound(n)%region%name
                     bound_id = bound(n)%id
! 
! .................. Listing by material number 
! 
                     IF(media_by_number) THEN
! 
                        n_t = Num_HetRegions + n
                        N_0 = n_t / 1000
                        N_1 = MOD(n_t,1000)
                        N_2 = N_1 /100
                        N_3 = MOD(N_1,100)
                        N_4 = N_3 / 10
                        N_5 = MOD(N_3,10)
! 
                        IF( medium(1:1) == 'F' .OR. medium(1:1) == 'f' ) THEN
                           medium = ' '//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        ELSE
                           medium = '-'//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        END IF
! 
                     END IF
! 
                  END IF IF_Range2
! 
! ............... Cartesian grid system, cylindrical-coordinate-defined boundaries : ERROR !!! 
! 
                  IF_Range3: IF( (bound(n)%region%coordinates(1:4) == 'CYLI' .OR. bound(n)%region%coordinates(1:4) == 'cyli' .OR. &
     &                            bound(n)%region%coordinates(1:4) == 'Cyli') )  &
     &            THEN
                     WRITE(*,6100) n,bound(n)%region%coordinates
                     STOP
                  END IF IF_Range3
! 
               END DO DO_Cartesian
! 
               RETURN
!
!***********************************************************************
!*                                                                     *
!*          For cylindrical coordinate systems                         *
!*                                                                     *
!***********************************************************************
!
            ELSE IF(coordinates(1:4) == 'CYLI' .OR. coordinates(1:4) == 'cyli' .OR. coordinates(1:4) == 'Cyli') THEN
! 
               DO_Cylindrical: DO n=1,Num_Boundaries
! 
! ............... Cylindrical grid system, cylindrical-coordinate-defined media boundaries
! 
                  IF_Range4: IF( (bound(n)%region%coordinates(1:4) == 'CYLI' .OR.    &
     &                            bound(n)%region%coordinates(1:4) == 'cyli' .OR.    &
     &                            bound(n)%region%coordinates(1:4) == 'Cyli') .AND.  &  
     &                            X >= bound(n)%region%L_min(1) .AND. X <= bound(n)%region%L_max(1) .AND.   &          
     &                            Z >= bound(n)%region%L_min(3) .AND. Z <= bound(n)%region%L_max(3)       ) & 
     &            THEN
!     
                     medium   = bound(n)%region%name
                     bound_id = bound(n)%id
! 
! .................. Listing by material number 
! 
                     IF(media_by_number) THEN
! 
                        n_t = Num_HetRegions + n
                        N_0 = n_t / 1000
                        N_1 = MOD(n_t,1000)
                        N_2 = N_1 /100
                        N_3 = MOD(N_1,100)
                        N_4 = N_3 / 10
                        N_5 = MOD(N_3,10)
! 
                        IF( medium(1:1) == 'F' .OR. medium(1:1) == 'f' ) THEN
                           medium = ' '//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        ELSE
                           medium = '-'//CHAR(N_0+48)//CHAR(N_2+48)//CHAR(N_4+48)//CHAR(N_5+48)
                        END IF
! 
                     END IF
! 
                  END IF IF_Range4
! 
! ............... Cylindrical grid system, other than cylindrical-coordinate-defined boundaries : ERROR !!! 
! 
                  IF_Range5: IF( (bound(n)%region%coordinates(1:4) == 'CART' .OR. bound(n)%region%coordinates(1:4) == 'cart' .OR. &
     &                            bound(n)%region%coordinates(1:4) == 'Cart') .OR.                                            &
     &                           (bound(n)%region%coordinates(1:4) == 'MIXE' .OR. bound(n)%region%coordinates(1:4) == 'mixe' .OR. &
     &                            bound(n)%region%coordinates(1:4) == 'Mixe') )  &
     &            THEN
! 
                     WRITE(*,6102) n,bound(n)%region%coordinates
                     STOP
                     RETURN
! 
                  END IF IF_Range5
! 
               END DO DO_Cylindrical
! 
               RETURN
! 
! 
! 
            END IF IF_Assign
!
!
!***********************************************************************
!*                                                                     *
!*                        F  O  R  M  A  T  S                          *
!*                                                                     *
!***********************************************************************
!
!
!6000 FORMAT(/,'Boundary_Status   1.0   31 August    2005',6X,'Determinimg whether a cell belongs to a boundary')
 6000 FORMAT(/,'Boundary_Status   1.0    3 August    2014',6X,'Determinimg whether a cell belongs to a boundary')
!
 6100 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The coordinate system of the grid is cylindrical, ', &
     &          'but the coordinates defining the boundary region #',I3,', are ',A11,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
 6102 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The coordinate system of the grid is Cartesian, ', &
     &          'but the coordinates defining the boundary region #',I3,', are ',A11,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
!
! =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>  End of Boundary_Status
!
!
            RETURN
!
         END SUBROUTINE Boundary_Status
!
!
!
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
! 
! 
! 
      END MODULE Boundary_Definition
!
!
!
!***********************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***********************************************************************
!
!
!
      MODULE Exclusion_Zone
! 
         SAVE
! 
! ----------
! ...... Derived-type variables     
! ----------
! 
         TYPE Excl_Zone
! 
            INTEGER :: num                                        ! Number of subregions in the exclusion zone 
! 
            REAL(KIND = 8), DIMENSION(:), POINTER :: X_min,X_max  ! Min and max extent of each of the subregions in ...
            REAL(KIND = 8), DIMENSION(:), POINTER :: Y_min,Y_max  ! ... the exclusion zone in the principal directions
            REAL(KIND = 8), DIMENSION(:), POINTER :: Z_min,Z_max  
            REAL(KIND = 8), DIMENSION(:), POINTER :: R_min,R_max  
! 
         END TYPE Excl_Zone
! 
         TYPE(Excl_Zone), ALLOCATABLE, DIMENSION(:) :: ExclZone
! 
! 
! 
      END MODULE Exclusion_Zone
!
!
!
!***********************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***********************************************************************
!
!
!
!     MODULE Grid_Generator
!
!        IMPLICIT NONE
! 
!        PRIVATE
! 
!        PUBLIC :: Mesh_Maker,MINC
!
!
! 
!        CONTAINS
!
!
!
!***********************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***********************************************************************
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE Mesh_Maker(Flag_MINC)
!
         USE MeshMaker_Data
         USE Grid_Gener_parameters
!
!***********************************************************************
!***********************************************************************
!*                                                                     *
!*            Executive routine for generating TOUGH+ grid             *
!*        and assigning initial property distribution by region        *
!*                                                                     *
!***********************************************************************
!***********************************************************************
!
      IMPLICIT NONE
! 
! -------
! ... Character variables     
! -------
! 
      CHARACTER(LEN = 5) :: word,KeyWord
      CHARACTER(LEN = 3) :: W_c1to3  
! 
! -------
! ... Logical variables     
! -------
! 
      LOGICAL :: exists
      LOGICAL, INTENT(OUT) :: Flag_MINC
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>  Main body of Mesh_Maker
!
!
      WRITE(*,6000)
! 
! ----------
! ... Initialization
! ----------
! 
      Flag_MINC = .FALSE.        !  Indicates non-MINC type of MESH
! 
      Flag_HetRegions = .FALSE.  !  Indicates uniform geologic media distribution
      Flag_Boundaries = .FALSE.  !  Indicates no-flow boundaries
      Flag_ExclZones  = .FALSE.  !  Indicates regularly-shaped domain
! 
! 
!***********************************************************************
!*                                                                     *
!*                 READING AND SEARCHING FOR KEYWORDS                  *
!*                                                                     *
!***********************************************************************
! 
! 
 1000 READ(*, FMT = 5001) word
! 
      KeyWord = TRIM(ADJUSTL(word))
      W_c1to3 = KeyWord(1:3)
! 
! 
! 
      CASE_KeyWord: SELECT CASE(W_c1to3)
! 
! 
!***********************************************************************
!*                                                                     *
!*    KeyWord = 'RZ2D': GENERATE CYLINDRICAL GRID                      *
!*                                                                     *
!***********************************************************************
! 
! 
      CASE ('RZ2','rz2')
!
         WRITE(*,6012) 
!
         CALL RZ2D
!
         IF(KeyWord == 'RZ2DL' .OR. KeyWord == 'rz2dl') THEN
            CALL WRZ2D(2)                                      ! Grid by layers
         ELSE
            CALL WRZ2D(1)                                      ! Grid by columns
         END IF
! 
! 
!***********************************************************************
!*                                                                     *
!*    KeyWord = 'XYZ': GENERATE CARTESIAN GRID                         *
!*                                                                     *
!***********************************************************************
! 
!
      CASE ('XYZ','xyz')
!
         WRITE(*,6014) 
!
         CALL GXYZ
! 
! 
!***********************************************************************
!*                                                                     *
!*    KeyWord = 'HET': Assign heterogeneous property distributions     *
!*                     by region (geometry-based)                      *
!*                                                                     *
!***********************************************************************
! 
!
      CASE ('HET','Het','het','REG','Reg','reg')
!
         WRITE(*,6015) 
!
         CALL Heterogns_Regions
!
         Flag_HetRegions = .TRUE.  !  Indicates heterogeneous distribution of geologic media 
!
         GO TO 1000
! 
! 
!***********************************************************************
!*                                                                     *
!*    KeyWord = 'EXC': Assign exclusion zones of the original grid     *
!*                     Defining irregularly-shaped domains             *
!*                                                                     *
!***********************************************************************
! 
!
      CASE ('EXC','Exc','exc')
!
         WRITE(*,6016) 
!
         Flag_ExclZones = .TRUE.  !  Indicates irregularly-shaped domain
!
         GO TO 1000
! 
! 
!***********************************************************************
!*                                                                     *
!*    KeyWord = 'BOU': Assign boundary subdomains                      *
!*                                                                     *
!***********************************************************************
! 
!
      CASE ('BOU','Bou','bou')
!
         WRITE(*,6018) 
!
         CALL Boundaries
!
         Flag_Boundaries = .TRUE.  !  Indicates boundaries of time-variant or invariant conditions
!
         GO TO 1000
! 
! 
! 
!***********************************************************************
!*                                                                     *
!*    KeyWord = 'MINC': GENERATE GRID FOR FRACTURED MEDIA              *
!*                                                                     *
!***********************************************************************
! 
!
      CASE ('MIN','Min','min')
! 
! -------------
! ...... Misspelled keyword (intended MINC)
! -------------
! 
         IF(KeyWord(4:4) /= 'C' .AND. KeyWord(4:4) /= 'c') THEN
! 
             WRITE(*,6100)
             STOP 
! 
! -------------
! ...... Correctly spelled keyword MINC
! -------------
! 
         ELSE
!
            Flag_MINC = .TRUE.
!
            INQUIRE(FILE='MINC', EXIST = exists)
!
            IF(exists) THEN
               WRITE(*,6021)
               OPEN(UNIT = MINC_Unit, FILE = 'MINC', STATUS = 'OLD')
            ELSE
               WRITE(*,6022)
               OPEN(UNIT = MINC_Unit, FILE = 'MINC', STATUS = 'NEW')
            END IF
!
            CALL MINC
!
         END IF
! 
! 
!***********************************************************************
!*                                                                     *
!*    Default case: Unknown keyword - The simulation is aborted        *
!*                                                                     *
!***********************************************************************
! 
! 
      CASE DEFAULT
! 
! ...... Printing the unknown keyword, and then stopping 
! 
         WRITE(*,6100) KeyWord
         STOP
!
!
! 
      END SELECT CASE_KeyWord
!
!
!***********************************************************************
!*                                                                     *
!*                        F  O  R  M  A  T  S                          *
!*                                                                     *
!***********************************************************************
!
!
!6000 FORMAT(/,'Mesh_Maker        1.0   29 August    2004',6X,'EXECUTIVE ROUTINE FOR INTERNAL MESH GENERATION')
 6000 FORMAT(/,'Mesh_Maker        1.5    8 August    2014',6X,'EXECUTIVE ROUTINE FOR INTERNAL MESH GENERATION')
!
 5001 FORMAT(A5)
!
 6012 FORMAT(/,' ',131('*'),   &
     &       /,' *',20X,'"Mesh_Maker" - RZ2D: GENERATE 2-D R-Z MESH',T132,'*',   &
     &       /,' ',131('*'))
 6014 format(/,' ',131('*'),   &
     &       /,' *',20X,'"Mesh_Maker" - XYZ: GENERATE 1, 2, OR 3-D CARTESIAN MESH',T132,'*',   &
     &       /,' ',131('*'))
 6015 format(/,' ',131('*'),   &
     &       /,' *',20X,'"Mesh_Maker" - HET: Define regions of heterogeneous media in the grid',T132,'*',   &
     &       /,' ',131('*'))
 6016 format(/,' ',131('*'),   &
     &       /,' *',20X,'"Mesh_Maker" - EXC: Describe irregularly-shaped grid by defining grid exclusion zones',T132,'*',   &
     &       /,' ',131('*'))
 6018 format(/,' ',131('*'),   &
     &       /,' *',20X,'"Mesh_Maker" - BOU: Define boundary regions/subdomains of the grid',T132,'*',   &
     &       /,' ',131('*'))
!
 6020 format(/,' ',131('*'),   &
     &       /,' *',18X,'"Mesh_Maker" - MINC: GENERATE MULTIPLE INTERACTING CONTINUA MESH FOR FRACTURED MEDIUM',T132,'*',   &
     &       /,' ',131('*')/)
!
 6021 FORMAT(' FILE "MINC" EXISTS ==> OPEN AS AN OLD FILE')
 6022 FORMAT(' FILE "MINC" DOES NOT EXIST ==> OPEN AS A NEW FILE')
!
 6024 FORMAT( /,' ',131('*'),   &
     &       //,' MESH GENERATION COMPLETE ==> EXIT FROM THE "Mesh_Maker" FACILITY',/)
!
 6100 FORMAT(' "Mesh_Maker" HAS READ THE UNKNOWN KEYWORD ',A5,'  ==> >>>>> STOP EXECUTION')
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>  End of Mesh_Maker
!
!
      RETURN
!
      END SUBROUTINE Mesh_Maker
!
!
!
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
!
!
!
         SUBROUTINE Heterogns_Regions
!
            USE Het_Region_Definition
            USE Grid_Gener_parameters, ONLY: dominant_medium
!
!***********************************************************************
!***********************************************************************
!*                                                                     *
!*    Routine defining regions of heterogeneous media in the grid      *
!*                                                                     *
!***********************************************************************
!***********************************************************************
!
            IMPLICIT NONE
! 
! -------------
! ......... Integer variables
! -------------
! 
            INTEGER :: ier,n_char,n
! 
! -------------
! ......... CHARACTER variables
! -------------
! 
            CHARACTER(LEN =  5) :: w_temp
!
!
! =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>  Main body of Heterogns_Regions
!
!
            WRITE(*,6000)
!
! ------------
! ......... Read the number of heterogeneous domains
! ------------
! 
            READ(*,*) Num_HetRegions
            IF( Num_HetRegions == 0 ) RETURN
!
! ------------
! ......... Allocate memory to the arrays defining the heterogeneous regions
! ------------
! 
            ALLOCATE(H_Region(Num_HetRegions), STAT=ier)
! 
! ......... Print explanatory comments  
! 
            IF(ier == 0) THEN
               WRITE(*,6101) 
            ELSE
               WRITE(*,6102) 
               WRITE(*,6102) 
               STOP
            END IF
! 
! -------------
! ......... Initialization 
! -------------
! 
            H_Region(1:Num_HetRegions) = Het_Region( "     ", "cartesian", "m",    &
      &                                              (/ 0.0d0,0.0d0,0.0d0 /),         &
      &                                              (/ 0.0d0,0.0d0,0.0d0 /), 0.0d0,0.0d0,0.0d0,0.0d0,0.0d0 )
!
!***********************************************************************
!*                                                                     *
!*      First medium is the dominant one - Only the name is read       *
!*                                                                     *
!***********************************************************************
!
            IF_First: IF(Num_HetRegions >= 1) THEN
!
               READ(*,*) dominant_medium
! 
! ............ Rewriting the medium number or name to be left-justified 
! 
               w_temp = '     '
               n_char = LEN_TRIM(ADJUSTL(dominant_medium))
!
               w_temp(1:n_char) = TRIM(ADJUSTL(dominant_medium))
               dominant_medium  = w_temp
!
               IF(Num_HetRegions == 1) RETURN
!
            END IF IF_First
!
!***********************************************************************
!*                                                                     *
!*         Read the specifics of the remaining media regions           *
!*                                                                     *
!***********************************************************************
!
            DO_Regions: DO n = 2,Num_HetRegions
! 
! -----------------
! ............ Read heterogeneous region specifics for Num_HetRegions > 1 
! -----------------
! 
               READ(*,*) H_Region(n)%name  
! 
! ............ Rewriting the medium number or name to be left-justified 
! 
               w_temp = '     '
               n_char = LEN_TRIM(ADJUSTL(H_Region(n)%name))
! 
               w_temp(1:n_char) = TRIM(ADJUSTL(H_Region(n)%name))
               H_Region(n)%name = w_temp
! 
! ----------------
! ............ Reading the type of coordinates 
! ----------------
! 
               READ(*,*) H_Region(n)%coordinates, H_Region(n)%units    
! 
! ............ Cartesian coordinates
! 
               IF(H_Region(n)%coordinates(1:4) == 'CART' .OR. H_Region(n)%coordinates(1:4) == 'cart' .OR. &
      &           H_Region(n)%coordinates(1:4) == 'Cart') &
      &        THEN
! 
                  READ(*,*) H_Region(n)%L_min(1),H_Region(n)%L_max(1), &
      &                     H_Region(n)%L_min(2),H_Region(n)%L_max(2), &   
      &                     H_Region(n)%L_min(3),H_Region(n)%L_max(3)   
! 
! ............... Ensuring that the L_min of the range is not >= of the L_max
! 
                  IF( H_Region(n)%L_min(1) >= H_Region(n)%L_max(1) ) THEN
                     WRITE(*,6050) H_Region(n)%L_min(1), n, H_Region(n)%L_max(1)
                     STOP
                  END IF
! 
                  IF( H_Region(n)%L_min(2) >= H_Region(n)%L_max(2) ) THEN
                     WRITE(*,6051) H_Region(n)%L_min(2), n, H_Region(n)%L_max(2)
                     STOP
                  END IF
! 
                  IF( H_Region(n)%L_min(3) >= H_Region(n)%L_max(3) ) THEN
                     WRITE(*,6052)  H_Region(n)%L_min(3), n, H_Region(n)%L_max(3)
                     STOP
                  END IF
! 
! ............ Cylindrical coordinates
! 
               ELSE IF(H_Region(n)%coordinates(1:4) == 'CYLI' .OR. H_Region(n)%coordinates(1:4) == 'cyli' .OR. &
      &                H_Region(n)%coordinates(1:4) == 'Cyli') &
      &        THEN
! 
                  READ(*,*) H_Region(n)%L_min(1),H_Region(n)%L_max(1), &
      &                     H_Region(n)%L_min(3),H_Region(n)%L_max(3)   
! 
! ............... Ensuring that the L_min of the range is not >= of the L_max
! 
                  IF( H_Region(n)%L_min(1) >= H_Region(n)%L_max(1) ) THEN
                     WRITE(*,6053) H_Region(n)%L_min(1), n, H_Region(n)%L_max(1)
                     STOP
                  END IF
! 
                  IF( H_Region(n)%L_min(3) >= H_Region(n)%L_max(3) ) THEN
                     WRITE(*,6052)  H_Region(n)%L_min(3), n, H_Region(n)%L_max(3)
                     STOP
                  END IF
! 
!                 READ(*,*) H_Region(n)%x0,H_Region(n)%y0,H_Region(n)%z0,H_Region(n)%R_min,H_Region(n)%R_max   
! 
! ............... Ensuring that the R_min of the range is not >= of the R_max
! 
!                 IF( H_Region(n)%R_min >= H_Region(n)%R_max ) THEN
!                    WRITE(*,6053) H_Region(n)%R_min, n, H_Region(n)%R_max
!                    STOP
!                 END IF
! 
! ............ Mixed coordinates
! 
               ELSE IF(H_Region(n)%coordinates(1:4) == 'MIXE' .OR. H_Region(n)%coordinates(1:4) == 'mixe' .OR. &
      &                H_Region(n)%coordinates(1:4) == 'Mixe') &
      &        THEN
! 
                  READ(*,*) H_Region(n)%L_min(1),H_Region(n)%L_max(1), &
      &                     H_Region(n)%L_min(2),H_Region(n)%L_max(2), &   
      &                     H_Region(n)%L_min(3),H_Region(n)%L_max(3), &   
      &                     H_Region(n)%x0,H_Region(n)%y0,H_Region(n)%z0,H_Region(n)%R_min,H_Region(n)%R_max   
! 
! ............... Ensuring that the L_min of the range is not >= of the L_max
! 
                  IF( H_Region(n)%L_min(1) >= H_Region(n)%L_max(1) ) THEN
                     WRITE(*,6050) H_Region(n)%L_min(1), n, H_Region(n)%L_max(1)
                     STOP
                  END IF
! 
                  IF( H_Region(n)%L_min(2) >= H_Region(n)%L_max(2) ) THEN
                     WRITE(*,6051) H_Region(n)%L_min(2), n, H_Region(n)%L_max(2)
                     STOP
                  END IF
! 
                  IF( H_Region(n)%L_min(3) >= H_Region(n)%L_max(3) ) THEN
                     WRITE(*,6052)  H_Region(n)%L_min(3), n, H_Region(n)%L_max(3)
                     STOP
                  END IF
! 
! ............... Ensuring that the R_min of the range is not >= of the R_max
! 
                  IF( H_Region(n)%R_min >= H_Region(n)%R_max ) THEN
                     WRITE(*,6053) H_Region(n)%R_min, n, H_Region(n)%R_max
                     STOP
                  END IF
! 
               ELSE 
! 
! ............ Unavailable/wrong coordinates
! 
                  WRITE(*,6105) H_Region(n)%coordinates   ! Print a clarifying message and stop
                  STOP
! 
               END IF
! 
! 
! 
            END DO DO_Regions
!
!***********************************************************************
!*                                                                     *
!*          Convert the region ranges into SI units (m)                *
!*                                                                     *
!***********************************************************************
!
! ......... Conversion into METERS from INCHES
!
            FORALL (n=2:Num_HetRegions, H_Region(n)%units == 'IN' .OR. H_Region(n)%units == 'in' .OR. H_Region(n)%units == 'In') 
               H_Region(n)%L_min(1:3) = H_Region(n)%L_min(1:3)*2.54d-2
               H_Region(n)%L_max(1:3) = H_Region(n)%L_max(1:3)*2.54d-2
               H_Region(n)%x0         = H_Region(n)%x0*2.54d-2
               H_Region(n)%y0         = H_Region(n)%y0*2.54d-2
               H_Region(n)%z0         = H_Region(n)%z0*2.54d-2
               H_Region(n)%R_min      = H_Region(n)%R_min*2.54d-2
               H_Region(n)%R_max      = H_Region(n)%R_max*2.54d-2  
            END FORALL
!
! ......... Conversion into METERS from FEET
!
            FORALL (n=2:Num_HetRegions, H_Region(n)%units == 'FT' .OR. H_Region(n)%units == 'ft' .OR. H_Region(n)%units == 'Ft') 
               H_Region(n)%L_min(1:3) = H_Region(n)%L_min(1:3)*3.038d-1
               H_Region(n)%L_max(1:3) = H_Region(n)%L_max(1:3)*3.038d-1
               H_Region(n)%x0         = H_Region(n)%x0*3.038d-1
               H_Region(n)%y0         = H_Region(n)%y0*3.038d-1
               H_Region(n)%z0         = H_Region(n)%z0*3.038d-1
               H_Region(n)%R_min      = H_Region(n)%R_min*3.038d-1
               H_Region(n)%R_max      = H_Region(n)%R_max*3.038d-1  
            END FORALL
!
! ......... Conversion into METERS from KM
!
            FORALL (n=2:Num_HetRegions, H_Region(n)%units == 'KM' .OR. H_Region(n)%units == 'km' .OR. H_Region(n)%units == 'Km') 
               H_Region(n)%L_min(1:3) = H_Region(n)%L_min(1:3)*1.0d3
               H_Region(n)%L_max(1:3) = H_Region(n)%L_max(1:3)*1.0d3
               H_Region(n)%x0         = H_Region(n)%x0*1.0d3
               H_Region(n)%y0         = H_Region(n)%y0*1.0d3
               H_Region(n)%z0         = H_Region(n)%z0*1.0d3
               H_Region(n)%R_min      = H_Region(n)%R_min*1.0d3
               H_Region(n)%R_max      = H_Region(n)%R_max*1.0d3  
            END FORALL
!
! ......... Conversion into METERS from MM
!
            FORALL (n=2:Num_HetRegions, H_Region(n)%units == 'MM' .OR. H_Region(n)%units == 'mm' .OR. H_Region(n)%units == 'Mm') 
               H_Region(n)%L_min(1:3) = H_Region(n)%L_min(1:3)*1.0d-3
               H_Region(n)%L_max(1:3) = H_Region(n)%L_max(1:3)*1.0d-3
               H_Region(n)%x0         = H_Region(n)%x0*1.0d-3
               H_Region(n)%y0         = H_Region(n)%y0*1.0d-3
               H_Region(n)%z0         = H_Region(n)%z0*1.0d-3
               H_Region(n)%R_min      = H_Region(n)%R_min*1.0d-3
               H_Region(n)%R_max      = H_Region(n)%R_max*1.0d-3  
            END FORALL
!
!
!***********************************************************************
!*                                                                     *
!*                        F  O  R  M  A  T  S                          *
!*                                                                     *
!***********************************************************************
!
!
 6000 FORMAT(/,'Heterogns_Regions 1.5   12 August    2014',6X,'Defining regions of heterogeneous media in the grid')
!
 6050 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The input X_min =',1pe12.5,' of the medium region # ',I2,  &
     &          ' is larger than (or equal to) the X_max =',1pe12.5,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
 6051 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The input Y_min =',1pe12.5,' of the medium region # ',I2,  &
     &          ' is larger than (or equal to) the Y_max =',1pe12.5,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
 6052 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The input Z_min =',1pe12.5,' of the medium region # ',I2,  &
     &          ' is larger than (or equal to) the Z_max =',1pe12.5,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
 6053 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The input R_min =',1pe12.5,' of the medium region # ',I2,  &
     &          ' is larger than (or equal to) the R_max =',1pe12.5,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
 6101 FORMAT(T5,'Memory allocation to array "H_Region" in subroutine "Heterogns_Regions" was successful')
 6102 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'Memory allocation to array "H_Region" in subroutine "Heterogns_Regions" was unsuccessful',//,   &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
 6105 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The coordinate system (read by "coordinates" = ',a11,' in subroutine "Heterogns_Regions") is unavailable',//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
!
! =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>  End of Heterogns_Regions
!
!
            RETURN
!
         END SUBROUTINE Heterogns_Regions
!
!
!
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
!
!
!
         SUBROUTINE Boundaries
!
            USE Boundary_Definition
!
!***********************************************************************
!***********************************************************************
!*                                                                     *
!*    Routine defining regions of heterogeneous media in the grid      *
!*                                                                     *
!***********************************************************************
!***********************************************************************
!
            IMPLICIT NONE
! 
! -------------
! ......... Integer variables
! -------------
! 
            INTEGER :: ier,n_char,n
! 
! -------------
! ......... CHARACTER variables
! -------------
! 
            CHARACTER(LEN =  5) :: w_temp
!
!
! =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   Main body of Boundaries
!
!
            WRITE(*,6000)
!
! ------------
! ......... Read the number of boundaries
! ------------
! 
            READ(*,*) Num_Boundaries
            IF( Num_Boundaries == 0 ) RETURN
!
! ------------
! ......... Allocate memory to the arrays defining the boundary regions
! ------------
! 
            ALLOCATE(bound(Num_Boundaries), STAT=ier)
! 
! ......... Print explanatory comments  
! 
            IF(ier == 0) THEN
               WRITE(*,6101) 
            ELSE
               WRITE(*,6102) 
               WRITE(*,6102) 
               STOP
            END IF
! 
! -------------
! ......... Initialization 
! -------------
! 
            bound(1:Num_Boundaries) = Boundary(" ", Het_Region( "     ", "cartesian", "m",       &
      &                                                         (/ 0.0d0,0.0d0,0.0d0 /),         &
      &                                                         (/ 0.0d0,0.0d0,0.0d0 /), 0.0d0,0.0d0,0.0d0,0.0d0,0.0d0  &
      &                                                       ) &
      &                                       )
!
!***********************************************************************
!*                                                                     *
!*             Read the specifics of the boundary regions              *
!*                                                                     *
!***********************************************************************
!
            DO_Regions: DO n = 1,Num_Boundaries
! 
! -----------------
! ............ Read boundary specifics 
! -----------------
! 
               READ(*,*) bound(n)%id,bound(n)%region%name  
! 
! ............ Resetting the boundary ID to the standard inactive status 
! 
               IF(bound(n)%id /= 'V') bound(n)%id = 'I'
! 
! ............ Rewriting the boundary medium number or name to be left-justified 
! 
               w_temp = '     '
               n_char = LEN_TRIM(ADJUSTL(bound(n)%region%name))
! 
               w_temp(1:n_char)     = TRIM(ADJUSTL(bound(n)%region%name))
               bound(n)%region%name = w_temp
! 
! ----------------
! ............ Reading the type of coordinates 
! ----------------
! 
               READ(*,*) bound(n)%region%coordinates, bound(n)%region%units    
! 
! ............ Cartesian coordinates
! 
               IF(bound(n)%region%coordinates(1:4) == 'CART' .OR. bound(n)%region%coordinates(1:4) == 'cart' .OR. &
      &           bound(n)%region%coordinates(1:4) == 'Cart') &
      &        THEN
! 
                  READ(*,*) bound(n)%region%L_min(1),bound(n)%region%L_max(1), &
      &                     bound(n)%region%L_min(2),bound(n)%region%L_max(2), &   
      &                     bound(n)%region%L_min(3),bound(n)%region%L_max(3)   
! 
! ............... Ensuring that the L_min of the range is not >= of the L_max
! 
                  IF( bound(n)%region%L_min(1) >= bound(n)%region%L_max(1) ) THEN
                     WRITE(*,6050) bound(n)%region%L_min(1), n, bound(n)%region%L_max(1)
                     STOP
                  END IF
! 
                  IF( bound(n)%region%L_min(2) >= bound(n)%region%L_max(2) ) THEN
                     WRITE(*,6051) bound(n)%region%L_min(2), n, bound(n)%region%L_max(2)
                     STOP
                  END IF
! 
                  IF( bound(n)%region%L_min(3) >= bound(n)%region%L_max(3) ) THEN
                     WRITE(*,6052)  bound(n)%region%L_min(3), n, bound(n)%region%L_max(3)
                     STOP
                  END IF
! 
! ............ Cylindrical coordinates
! 
               ELSE IF(bound(n)%region%coordinates(1:4) == 'CYLI' .OR. bound(n)%region%coordinates(1:4) == 'cyli' .OR. &
      &                bound(n)%region%coordinates(1:4) == 'Cyli') &
      &        THEN
! 
                  READ(*,*) bound(n)%region%L_min(1),bound(n)%region%L_max(1), &
      &                     bound(n)%region%L_min(3),bound(n)%region%L_max(3)   
! 
! ............... Ensuring that the L_min of the range is not >= of the L_max
! 
                  IF( bound(n)%region%L_min(1) >= bound(n)%region%L_max(1) ) THEN
                     WRITE(*,6053) bound(n)%region%L_min(1), n, bound(n)%region%L_max(1)
                     STOP
                  END IF
! 
                  IF( bound(n)%region%L_min(3) >= bound(n)%region%L_max(3) ) THEN
                     WRITE(*,6052)  bound(n)%region%L_min(3), n, bound(n)%region%L_max(3)
                     STOP
                  END IF
! 
! ............ Mixed coordinates
! 
               ELSE IF(bound(n)%region%coordinates(1:4) == 'MIXE' .OR. bound(n)%region%coordinates(1:4) == 'mixe' .OR. &
      &                bound(n)%region%coordinates(1:4) == 'Mixe') &
      &        THEN
! 
                  READ(*,*) bound(n)%region%L_min(1),bound(n)%region%L_max(1),        &
      &                     bound(n)%region%L_min(2),bound(n)%region%L_max(2),        &   
      &                     bound(n)%region%L_min(3),bound(n)%region%L_max(3),        &   
      &                     bound(n)%region%x0,bound(n)%region%y0,bound(n)%region%z0, &
      &                     bound(n)%region%R_min,bound(n)%region%R_max   
! 
! ............... Ensuring that the L_min of the range is not >= of the L_max
! 
                  IF( bound(n)%region%L_min(1) >= bound(n)%region%L_max(1) ) THEN
                     WRITE(*,6050) bound(n)%region%L_min(1), n, bound(n)%region%L_max(1)
                     STOP
                  END IF
! 
                  IF( bound(n)%region%L_min(2) >= bound(n)%region%L_max(2) ) THEN
                     WRITE(*,6051) bound(n)%region%L_min(2), n, bound(n)%region%L_max(2)
                     STOP
                  END IF
! 
                  IF( bound(n)%region%L_min(3) >= bound(n)%region%L_max(3) ) THEN
                     WRITE(*,6052)  bound(n)%region%L_min(3), n, bound(n)%region%L_max(3)
                     STOP
                  END IF
! 
! ............... Ensuring that the R_min of the range is not >= of the R_max
! 
                  IF( bound(n)%region%R_min >= bound(n)%region%R_max ) THEN
                     WRITE(*,6053) bound(n)%region%R_min, n, bound(n)%region%R_max
                     STOP
                  END IF
! 
               ELSE 
! 
! ............ Unavailable/wrong coordinates
! 
                  WRITE(*,6105) bound(n)%region%coordinates   ! Print a clarifying message and stop
                  STOP
! 
               END IF
! 
! 
! 
            END DO DO_Regions
!
!***********************************************************************
!*                                                                     *
!*          Convert the boundary ranges into SI units (m)              *
!*                                                                     *
!***********************************************************************
!
! ......... Conversion into METERS from INCHES
!
            FORALL (n=1:Num_Boundaries, bound(n)%region%units == 'IN' .OR. bound(n)%region%units == 'in' .OR. &
     &                                  bound(n)%region%units == 'In') 
               bound(n)%region%L_min(1:3) = bound(n)%region%L_min(1:3)*2.54d-2
               bound(n)%region%L_max(1:3) = bound(n)%region%L_max(1:3)*2.54d-2
               bound(n)%region%x0         = bound(n)%region%x0*2.54d-2
               bound(n)%region%y0         = bound(n)%region%y0*2.54d-2
               bound(n)%region%z0         = bound(n)%region%z0*2.54d-2
               bound(n)%region%R_min      = bound(n)%region%R_min*2.54d-2
               bound(n)%region%R_max      = bound(n)%region%R_max*2.54d-2  
!
            END FORALL
!
! ......... Conversion into METERS from FEET
!
            FORALL (n=1:Num_Boundaries, bound(n)%region%units == 'FT' .OR. bound(n)%region%units == 'ft' .OR. & 
     &                                  bound(n)%region%units == 'Ft') 
               bound(n)%region%L_min(1:3) = bound(n)%region%L_min(1:3)*3.038d-1
               bound(n)%region%L_max(1:3) = bound(n)%region%L_max(1:3)*3.038d-1
               bound(n)%region%x0         = bound(n)%region%x0*3.038d-1
               bound(n)%region%y0         = bound(n)%region%y0*3.038d-1
               bound(n)%region%z0         = bound(n)%region%z0*3.038d-1
               bound(n)%region%R_min      = bound(n)%region%R_min*3.038d-1
               bound(n)%region%R_max      = bound(n)%region%R_max*3.038d-1  
!
            END FORALL
!
! ......... Conversion into METERS from KM
!
            FORALL (n=1:Num_Boundaries, bound(n)%region%units == 'KM' .OR. bound(n)%region%units == 'km' .OR. & 
     &                                  bound(n)%region%units == 'Km') 
               bound(n)%region%L_min(1:3) = bound(n)%region%L_min(1:3)*1.0d3
               bound(n)%region%L_max(1:3) = bound(n)%region%L_max(1:3)*1.0d3
               bound(n)%region%x0         = bound(n)%region%x0*1.0d3
               bound(n)%region%y0         = bound(n)%region%y0*1.0d3
               bound(n)%region%z0         = bound(n)%region%z0*1.0d3
               bound(n)%region%R_min      = bound(n)%region%R_min*1.0d3
               bound(n)%region%R_max      = bound(n)%region%R_max*1.0d3  
!
            END FORALL
!
! ......... Conversion into METERS from MM
!
            FORALL (n=1:Num_Boundaries, bound(n)%region%units == 'MM' .OR. bound(n)%region%units == 'mm' .OR.  &
     &                                  bound(n)%region%units == 'Mm') 
               bound(n)%region%L_min(1:3) = bound(n)%region%L_min(1:3)*1.0d-3
               bound(n)%region%L_max(1:3) = bound(n)%region%L_max(1:3)*1.0d-3
               bound(n)%region%x0         = bound(n)%region%x0*1.0d-3
               bound(n)%region%y0         = bound(n)%region%y0*1.0d-3
               bound(n)%region%z0         = bound(n)%region%z0*1.0d-3
               bound(n)%region%R_min      = bound(n)%region%R_min*1.0d-3
               bound(n)%region%R_max      = bound(n)%region%R_max*1.0d-3  
!
            END FORALL
!
!
!***********************************************************************
!*                                                                     *
!*                        F  O  R  M  A  T  S                          *
!*                                                                     *
!***********************************************************************
!
!
 6000 FORMAT(/,'Boundaries        1.5    6 August    2014',6X,'Defining the boundaries (inactive elements) of the grid')
!
 6050 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The input X_min =',1pe12.5,' of the boundary region # ',I2,  &
     &          ' is larger than (or equal to) the X_max =',1pe12.5,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
 6051 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The input Y_min =',1pe12.5,' of the boundary region # ',I2,  &
     &          ' is larger than (or equal to) the Y_max =',1pe12.5,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
 6052 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The input Z_min =',1pe12.5,' of the boundary region # ',I2,  &
     &          ' is larger than (or equal to) the Z_max =',1pe12.5,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
 6053 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The input R_min =',1pe12.5,' of the boundary region # ',I2,  &
     &          ' is larger than (or equal to) the R_max =',1pe12.5,//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
 6101 FORMAT(T5,'Memory allocation to array "bound" in subroutine "Boundaries" was successful')
 6102 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'Memory allocation to array "bound" in subroutine "Boundaries" was unsuccessful',//,   &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
 6105 FORMAT(//,20('ERROR-'),//,   &
     &       T5,'The coordinate system (read by "coordinates" = ',a11,' in subroutine "Boundaries") is unavailable',//, &
     &       T5,'!!!!!!!    THE EXECUTION WAS STOPPED    !!!!!!!',   &
     &       //,20('ERROR-'))
!
!
! =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of Boundaries
!
!
            RETURN
!
         END SUBROUTINE Boundaries
!
!
!
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
! >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE RZ2D
! 
         USE MeshMaker_Data
         USE Grid_Gener_parameters
! 
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
! 
      CHARACTER SUB*5,WORD*5
!
      COMMON/W1/V(2501),A(2501),D1(2501),D2(2501),RC(2501),H(200)
! 
      DIMENSION SUB(4)
!
      SAVE ICALL,SUB
!
      DATA SUB/'RADII','EQUID','LOGAR','LAYER'/
      DATA ICALL/0/
!
      S(X)=X*(X**NLOG-1.)/(X-1.0d0)
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   Main body of RZ2D
!
!
      ICALL=ICALL+1
      IF(ICALL == 1) WRITE(*,899)
!
! 899 FORMAT('RZ2D              1.0   11 FEBRUARY  1992',6X,'CALCULATE 2-D R-Z MESH FROM INPUT DATA')
  899 FORMAT('RZ2D              1.5   24 September 2013',6X,'CALCULATE 2-D R-Z MESH FROM INPUT DATA')
!
      NRAD=0
      NEQU=0
      NLOG=0
      NRADC=0
      NEQUC=0
      NLOGC=0
      NINT=0
!
  204 READ 200,WORD
  200 FORMAT(A5)
!
      DO 201 K=1,4
      IF(WORD.EQ.SUB(K)) GOTO 202
  201 CONTINUE
!
      PRINT 203,WORD
  203 FORMAT(' HAVE READ UNKNOWN BLOCK LABEL ',A5,'  ---  STOP EXECUTION')
      STOP
!
  202 GOTO(211,212,213,214),K
!
  211 CONTINUE
!-----READ A SET OF INTERFACE RADII.
      READ 220, NRAD
  220 FORMAT(I5,5X,7E10.4)
!
      NINT1=NINT+1
      NINT=NINT+NRAD
      NRADC=NRADC+NRAD
      READ 221,(RC(I),I=NINT1,NINT)
!
      RC(NINT1:NINT) = RC(NINT1:NINT) * factor
!
  221 FORMAT(8E10.4)
      GOTO 204
!
  212 CONTINUE
!-----READ DATA FOR GENERATING SOME EQUIDISTANT INTERFACES.
!
      READ 220,NEQU,DR
!
      DR = DR * factor
!
      NINT1=NINT+1
      NINT=NINT+NEQU
      NEQUC=NEQUC+NEQU
      DO 222 I=NINT1,NINT
  222 RC(I)=RC(I-1)+DR
      GOTO 204
!
  213 CONTINUE
!-----READ DATA FOR GENERATING INTERFACES WITH LOGARITHMICALLY
!     INCREASING DISTANCE.
!
      READ 220,NLOG,RLOG,DR
      NLOGC=NLOGC+NLOG
!
      DR   = DR * factor
      RLOG = RLOG * factor
!
      IF(DR.EQ.0.) DR = RC(NINT)-RC(NINT-1)
!
!
      PRINT 109
  109 FORMAT(/' ***** FIND APPROPRIATE FACTOR FOR INCREASING RADIAL DISTANCES *****')
!
      IX=0
      AM=(RLOG-RC(NINT))/DR
      AN=FLOAT(NLOG)
      XM=AM/AN
      IF(XM.EQ.1.) GO TO 40
      IF(AM.GT.AN) THEN
      XL=1.
      XR=(AM+1.)**(1./AN)
      XR=MIN(XR,XM)
      ELSE
      XL=AM/(AM+1.)
      XL=MAX(XL,XM)
      XR=1.
      ENDIF
!     FIND PROPER INCREMENT FACTOR THROUGH ITERATIVE BISECTING
   41 XM=0.5*(XL+XR)
      IF(XR.GT.2.*XL) XM=SQRT(XL*XR)
      SMID=S(XM)
      IX=IX+1
!     PRINT 42,IX,XL,XR,XM,SMID,AM
   42 FORMAT(' AT ITERATION',I3,'  XL=',E20.14,'  XR=',E20.14,   &
     &       '  XM=',E12.6,'  SMID=',E12.6,'  AM=',E12.6)
      IF(IX.GE.1000) GO TO 43
      IF((XR-XL) <= (XL*1.E-10)) GO TO 40
      IF(SMID <= AM) XL=XM
      IF(SMID >= AM) XR=XM
      GO TO 41
!
   40 RAT=XM
!
      PRINT 30,IX,RAT
   30 FORMAT(6X,' AFTER',I4,' ITERATIONS FOUND INCREMENT FACTOR RAT =',E14.8)
!
      DO I=1,NLOG
         DR         = DR*RAT
         RC(NINT+I) = RC(NINT+I-1)+DR
      END DO
!
      NINT = NINT+NLOG
      GO TO 204
!
  214 CONTINUE
!-----COME HERE AFTER ALL NINT INTERFACES ARE ASSIGNED, AND READ
!     DATA ON LAYER THICKNESSES.
!
      READ(*,*) NLAY, Z_ref
      READ 221,(H(I),I=1,NLAY)
!
      z_ref     = z_ref * factor
      H(1:NLAY) = H(1:NLAY) * factor
!
!
      DO I=2,NLAY
         IF(H(I) == 0.0d0) H(I) = H(I-1)
         IF(H(I) <  0.0d0) H(I) = 0.0d0
      END DO
!     
      IF(H(1) < 0.0d0) H(1) = 0.0d0
!
!-----NOW ASSIGN ALL GEOMETRY DATA FOR A LAYER OF UNIT THICKNESS.
!
!=====PRINT ALL GEOMETRIC QUANTITIES=====
!
!
      PRINT 19
   19 FORMAT(//' * * * * * M E S H  G E O M E T R Y * * * * *',10X,   &
     &'VOLUME AND AREA ARE GIVEN FOR HEIGHT = 1 METER'//   &
     &' ELEM     REL           RCON           D1             D2     ',   &
     &'       V              A'/)
!
      NELEMT=NINT-1
! 
! ----------------
! ........ KP - 2/11/92
! ........ Begin change to avoid "funny" multiply-add-related
! ........ behavior on IBM RISC6000
! ----------------
! 
      DO 15 I=1,NELEMT
         A(I)  = 2.0d0*PI*RC(I+1)
         REL   = SQRT(RC(I+1)*RC(I))
!
! ...... GJM - 1/18/08
!
!        REL = SQRT( 5.0d-1*(RC(I+1)*RC(I+1) - RC(I)*RC(I))/ DLOG((RC(I+1)/RC(I))))
         D1(I) = REL     - RC(I)
         D2(I) = RC(I+1) - REL
!        D(I)  = (RC(I+1) - RC(I))/2.0D0
         v(i)  = pi*(rc(i+1)+rc(i))*(rc(i+1)-rc(i))
   15 CONTINUE
!
      DO I=1,NELEMT
!        REL = RC(I+1) - D(I)
         REL = SQRT(RC(I+1)*RC(I))
!
! ...... GJM - 1/18/08
!
!        REL = SQRT( 5.0d-1*(RC(I+1)*RC(I+1) - RC(I)*RC(I))/ DLOG((RC(I+1)/RC(I))))
         PRINT 18, I,REL,RC(I+1),D1(I),D2(I),V(I),A(I)
      END DO
   18 FORMAT(' ',I4,6(2X,1PE12.5))
!
! 
! ----------------
! ........ KP - 2/11/92
! ........ End change 
! ----------------
! 
!
      REWIND (UNIT = MESH_Unit)
      PRINT 20
   20 FORMAT(/' WRITE FILE *MESH* FROM INPUT DATA')
! 
! - GJM: Begin modification for 8-character elements
! 
      IF_Format: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
         WRITE(MESH_Unit, FMT = '(T1, A)') '>>>ELEMENTS: Cylindrical coordinates'
         WRITE(MESH_Unit, FMT = '(T1, A)') '&Units  length_units = "m" /'
      ELSE
! 
         IF(ElemNameLength == 8) THEN
            WRITE(MESH_Unit,5003) NELEMT,NRADC,NEQUC,NLOGC,RC(1),RC(NELEMT+1)
         ELSE
            WRITE(MESH_Unit,3) NELEMT,NRADC,NEQUC,NLOGC,RC(1),RC(NELEMT+1)
         END IF
! 
      END IF IF_Format
! 
! 
 5003 FORMAT('ELEMEext2 --- ',4I5,F10.5,F10.3)
    3 FORMAT('ELEME --- ',4I5,F10.5,F10.3)
! 
!! - GJM: End modification for 8-character elements
! 
      RETURN
!
   43 CONTINUE
      PRINT 42,IX,XL,XR,XM,SMID,AM
      PRINT 44
   44 FORMAT(' CONVERGENCE FAILURE IN SUBROUTINE RZ2D',/,' STOP EXECUTION')
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of RZ2D
!
!
      STOP
      END SUBROUTINE RZ2D
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE WRZ2D(K)
! 
         USE Grid_Gener_parameters
         USE Het_Region_Definition
         USE Boundary_Definition
! 
         USE MeshMaker_Data
! 
! 
! 
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
! 
      CHARACTER NII*1,NII1*1,NOV*1,NOV1*1,DOM*5,bound_id*1,medium*5
! 
      COMMON/W1/V(2501),A(2501),D1(2501),D2(2501),RC(2501),H(200)
! 
      SAVE ICALL
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of WRZ2D
!
!
      ICALL=ICALL+1
      IF(ICALL == 1) WRITE(*,899)
!
! 899 FORMAT(6X,'WRZ2D    1.0      26 MARCH     1991',6X,
! 899 FORMAT('WRZ2D             1.1   20 September 2000',6X,'WRITE DATA FOR 2-D R-Z MESH ON FILE *MESH*')
  899 FORMAT('WRZ2D             1.5   11 August    2014',6X,'WRITE DATA FOR 2-D R-Z MESH ON FILE *MESH*')
!
      NSKIP1=1
      NSKIP2=2
!
!      NII=' '
      IF(K.EQ.2) GO TO 1
!
!-----COME HERE TO GENERATE MESH BY VERTICAL COLUMNS
!      DO 4 I= NSKIP1,NELEMT
!      MI = MOD(I,100)
      DO 4 I8=NSKIP1-1,NELEMT-1
         MI  = MOD(I8,100)
         I   = I8+1
!
         IF(MOD(I,100) == 0) THEN
            II = I/100
         ELSE
            II = I/100+1
         END IF
         NII = NA(II)
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ Begin change to assign correct location for average gridblock pressure
! ----------------
! 
!        REL = RC(I+1)-D(I)
         REL = SQRT(RC(I+1)*RC(I))
!         
! ...... GJM - 9/19/2000
!         
!        REL = SQRT( 5.0d-1*(RC(I+1)*RC(I+1) - RC(I)*RC(I))/ DLOG((RC(I+1)/RC(I))))
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ End change 
! ----------------
! 
         ZJ  = - H(1)/2.0d0
! 
! - GJM: Begin modification for 8-character elements
! 
         IF(ElemNameLength.EQ.8) THEN
            DO_JLoop1: DO J=1,NLAY
               MJ  = MOD(J-1,36)+1
               NOV = NA((J-1)/36+11)
               IF(J.GT.1) ZJ=ZJ-H(J)/2.-H(J-1)/2.0d0
               AIJ = 0.0d0
               IF(J.EQ.1) AIJ = AIJ+V(I)
               IF(J.EQ.NLAY) AIJ = AIJ+V(I)
               VIJ = V(I)*H(J)
!
! ----------------
! ............ When the domain is homogeneous, assign the standard domain number/name
! ----------------
! 
               IF_Media1: IF(Flag_HetRegions .EQV. .FALSE.) THEN
! 
                  DOM = '    1'
!
! ----------------
! ............ When the domain is heterogeneous ...
! ----------------
! 
               ELSE 
! 
! ............... Assign the dominant domain number/name
! 
                  IF_HetRegions1: IF(Num_HetRegions == 1) THEN
! 
                     DOM = dominant_medium
! 
! ............... Assign the appropriate heterogeneous domain number/name
! 
                  ELSE
                     DOM = Media_Region(Flag_HetRegions,'Cylindrical',dominant_medium,REL,Z_ref+ZJ,Z_ref+ZJ, media_by_number)
                  END IF IF_HetRegions1
! 
               END IF IF_Media1 
!
! ----------------
! ............ Determine whether this is a boundary cell
! ----------------
! 
               IF(Flag_Boundaries .AND. Num_Boundaries > 0) THEN 
! 
                  CALL Boundary_Status('Cylindrical',REL,Z_ref+ZJ,Z_ref+ZJ,bound_id,medium,media_by_number)
! 
                  IF(bound_id == 'I' .OR. bound_id == 'V') THEN
                     DOM = medium
                  END IF
! 
               ELSE
                  bound_id = ' '
               END IF
! 
 5401 FORMAT('ZZZ',3A1,I2.2,'  &Elem  V=',1ES17.10,', MedNum=',A,', r=',1ES17.10,', z=',1ES17.10,', act="',A,'" /')
 5400 FORMAT('ZZZ',3A1,I2.2,'  &Elem  V=',1ES17.10,', MedNum=',A,', r=',1ES17.10,', z=',1ES17.10,' /')
! 
 5501 FORMAT('ZZZ',3A1,I2.2,'  &Elem  V=',1ES17.10,', MedNam="',A,'", r=',1ES17.10,', z=',1ES17.10,', act="',A,'" /')
 5500 FORMAT('ZZZ',3A1,I2.2,'  &Elem  V=',1ES17.10,', MedNam="',A,'", r=',1ES17.10,', z=',1ES17.10,' /')
! 
               IF_Format: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
                  IF(bound_id == ' ') THEN
                     IF(DOM == '    1') THEN  
                        WRITE(MESH_Unit, FMT = 5400) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ
                     ELSE 
                        WRITE(MESH_Unit, FMT = 5500) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ
                     END IF
                  ELSE
                     IF(DOM == '    1') THEN  
                        WRITE(MESH_Unit, FMT = 5401) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ,bound_id
                     ELSE
                        WRITE(MESH_Unit, FMT = 5501) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ,bound_id
                     END IF
                  END IF
               ELSE
                  WRITE(MESH_Unit,5005) 'ZZZ',NOV,NA(MJ),NII,MI,DOM,VIJ,AIJ,REL,Z_ref+ZJ,bound_id
               END IF IF_Format
! 
            END DO DO_JLoop1 
         ELSE 
            DO_JLoop2: DO J=1,NLAY
               MJ  = MOD(J-1,36)+1
               NOV = NA((J-1)/36+11)
               IF(J.GT.1) ZJ=ZJ-H(J)/2.-H(J-1)/2.0d0
               AIJ = 0.0d0
               IF(J.EQ.1) AIJ = AIJ+V(I)
               IF(J.EQ.NLAY) AIJ = AIJ+V(I)
               VIJ = V(I)*H(J)
!
! ----------------
! ............ When the domain is homogeneous, assign the standard domain number/name
! ----------------
! 
               IF_Media2: IF(Flag_HetRegions .EQV. .FALSE.) THEN
! 
                  DOM = '    1'
!
! ----------------
! ............ When the domain is heterogeneous ...
! ----------------
! 
               ELSE 
! 
! ............... Assign the dominant domain number/name
! 
                  IF_HetRegions2: IF(Num_HetRegions == 1) THEN
! 
                     DOM = dominant_medium
! 
! ............... Assign the appropriate heterogeneous domain number/name
! 
                  ELSE
                     DOM = Media_Region(Flag_HetRegions,'Cylindrical',dominant_medium,REL,Z_ref+ZJ,Z_ref+ZJ, media_by_number)
                  END IF IF_HetRegions2
! 
               END IF IF_Media2 
!
! ----------------
! ............ Determine whether this is a boundary cell
! ----------------
! 
               IF(Flag_Boundaries .AND. Num_Boundaries > 0) THEN 
! 
                  CALL Boundary_Status('Cylindrical',REL,Z_ref+ZJ,Z_ref+ZJ,bound_id,medium,media_by_number)
! 
                  IF(bound_id == 'I' .OR. bound_id == 'V') THEN
                     DOM = medium
                  END IF
! 
               ELSE
                  bound_id = ' '
               END IF
! 
 5403 FORMAT(3A1,I2.2,'  &Elem  V=',1ES17.10,', MedNum=',A,', r=',1ES17.10,', z=',1ES17.10,', act="',A,'" /')
 5402 FORMAT(3A1,I2.2,'  &Elem  V=',1ES17.10,', MedNum=',A,', r=',1ES17.10,', z=',1ES17.10,' /')
! 
 5503 FORMAT(3A1,I2.2,'  &Elem  V=',1ES17.10,', MedNam="',A,'", r=',1ES17.10,', z=',1ES17.10,', act="',A,'" /')
 5502 FORMAT(3A1,I2.2,'  &Elem  V=',1ES17.10,', MedNam="',A,'", r=',1ES17.10,', z=',1ES17.10,' /')
! 
               IF_Format2: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
                  IF(bound_id == ' ') THEN
                     IF(DOM == '    1') THEN  
                        WRITE(MESH_Unit, FMT = 5402) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ
                     ELSE
                        WRITE(MESH_Unit, FMT = 5502) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ
                     END IF
                  ELSE
                     IF(DOM == '    1') THEN  
                        WRITE(MESH_Unit, FMT = 5403) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ,bound_id
                     ELSE
                        WRITE(MESH_Unit, FMT = 5503) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ,bound_id
                     END IF
                  END IF
               ELSE
                  WRITE(MESH_Unit,5) NOV,NA(MJ),NII,MI,DOM,VIJ,AIJ,REL,Z_ref+ZJ,bound_id
               END IF IF_Format2
! 
            END DO DO_JLoop2
         END IF 
    4 CONTINUE
!
    5 FORMAT(   3A1,I2.2,10X,a5,2(1PE10.4),10X,1PE10.4,10X,1PE10.3,1X,A1)
 5005 FORMAT(A3,3A1,I2.2,10X,a5,2(1PE10.4),10X,F10.3,10X,F10.3,1X,A1)
! 
!! - GJM: End modification for 8-character elements
! 
!
!     MJ RUNS FROM 1 TO 36 FOR CONSECUTIVE LAYERS.
!     NOV RUNS THROUGH A, B, C, ... FOR THE 1ST, 2ND, 3RD, ... SET
!     OF 36 LAYERS.
!
!-----CONVENTION FOR ELEMENT NAMES 
!     NOV = A, B, C, .... LABELS 1ST, 2ND, 3RD, ... SET OF 37 LAYERS.
!     NA = 0, 1, 2, 3, ... 9, A, B, C, ... LABELS LAYERS.
!     NII = 0, 1, 2, 3, ... AFTER 100, 200, 300, ... RADIAL ELEMENTS.
!     MI = 0, 1, 2, 3, ... 99; NUMBERS RADIAL ELEMENTS IN EXCESS OF
!                              FULL HUNDREDS.
!
      GOTO 2
!
    1 CONTINUE
!-----COME HERE TO GENERATE MESH BY HORIZONTAL LAYERS
      ZJ=-H(1)/2.
!     DO_JLoop3: DO 3 J=1,NLAY
      DO 3 J=1,NLAY
         MJ  = MOD(J-1,36)+1
         NOV = NA((J-1)/36+11)
         IF(J > 1) ZJ=ZJ-H(J)/2.-H(J-1)/2.
!        DO 3 I = NSKIP1,NELEMT
         DO 3 I8 = NSKIP1-1,NELEMT-1
!           MI  = MOD(I,100)
            MI  = MOD(I8,100)
            I   = I8+1
!
            IF(MOD(I,100) == 0) THEN
               II = I/100
            ELSE
               II = I/100+1
            END IF
            NII = NA(II)
!           IF(II.GE.1) NII = NA(II)
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ Begin change to assign correct location for average gridblock pressure
! ----------------
! 
!           REL = RC(I+1)-D(I)
            REL = SQRT(RC(I+1)*RC(I))
!           REL = SQRT( 5.0d-1*(RC(I+1)*RC(I+1) - RC(I)*RC(I))/ DLOG((RC(I+1)/RC(I))))
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ End change
! ----------------
! 
            AIJ = 0.0d0
            IF(J == 1)    AIJ = AIJ+V(I)
            IF(J == NLAY) AIJ = AIJ+V(I)
            VIJ = V(I)*H(J)
! 
! - GJM: Begin modification for 8-character elements
! 
!
! ----------------
! ............ When the domain is homogeneous, assign the standard domain number/name
! ----------------
! 
               IF_Media3: IF(Flag_HetRegions .EQV. .FALSE.) THEN
! 
                  DOM = '    1'
!
! ----------------
! ............ When the domain is heterogeneous ...
! ----------------
! 
               ELSE 
! 
! ............... Assign the dominant domain number/name
! 
                  IF_HetRegions3: IF(Num_HetRegions == 1) THEN
! 
                     DOM = dominant_medium
! 
! ............... Assign the appropriate heterogeneous domain number/name
! 
                  ELSE
                     DOM = Media_Region(Flag_HetRegions,'Cylindrical',dominant_medium,REL,Z_ref+ZJ,Z_ref+ZJ, media_by_number)
                  END IF IF_HetRegions3
! 
               END IF IF_Media3 
!
! ----------------
! ............ Determine whether this is a boundary cell
! ----------------
! 
               IF(Flag_Boundaries .AND. Num_Boundaries > 0) THEN 
! 
                  CALL Boundary_Status('Cylindrical',REL,Z_ref+ZJ,Z_ref+ZJ,bound_id,medium,media_by_number)
! 
                  IF(bound_id == 'I' .OR. bound_id == 'V') THEN
                     DOM = medium
                  END IF
! 
               ELSE
                  bound_id = ' '
               END IF
! 
! 
! 
               IF_Format3: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
! 
                  IF(bound_id == ' ') THEN
                     IF(ElemNameLength.EQ.8) THEN
                        IF(DOM == '    1') THEN  
                           WRITE(MESH_Unit, FMT = 5400) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ
                        ELSE
                           WRITE(MESH_Unit, FMT = 5500) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ
                        END IF
                     ELSE
                        IF(DOM == '    1') THEN  
                           WRITE(MESH_Unit, FMT = 5402) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ
                        ELSE
                           WRITE(MESH_Unit, FMT = 5502) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ
                        END IF
                     END IF
                  ELSE
                     IF(ElemNameLength.EQ.8) THEN
                        IF(DOM == '    1') THEN  
                           WRITE(MESH_Unit, FMT = 5401) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ,bound_id
                        ELSE
                           WRITE(MESH_Unit, FMT = 5501) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ,bound_id
                        END IF
                     ELSE
                        IF(DOM == '    1') THEN  
                           WRITE(MESH_Unit, FMT = 5403) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ,bound_id
                        ELSE
                           WRITE(MESH_Unit, FMT = 5503) NOV,NA(MJ),NII,MI,VIJ,DOM,REL,Z_ref+ZJ,bound_id
                        END IF
                     END IF
                  END IF
! 
               ELSE
! 
                  IF(ElemNameLength.EQ.8) THEN
                     WRITE(MESH_Unit,5005) 'ZZZ',NOV,NA(MJ),NII,MI,DOM,VIJ,AIJ,REL,Z_ref+ZJ,bound_id
                  ELSE
                     WRITE(MESH_Unit,5) NOV,NA(MJ),NII,MI,DOM,VIJ,AIJ,REL,Z_ref+ZJ,bound_id
                  END IF
! 
               END IF IF_Format3
! 
! - GJM: End modification for 8-character elements
! 
    3 CONTINUE
!
!     MJ RUNS FROM 1 TO 35 FOR CONSECUTIVE LAYERS.
!     NOV RUNS THROUGH A, B, C, ... FOR THE 1ST, 2ND, 3RD, ... SET
!     OF 35 LAYERS.
!
    2 CONTINUE
!
!
      IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
         WRITE(MESH_Unit, FMT = 5551) 
       ELSE
         WRITE(MESH_Unit, FMT = 6)
      END IF 
!
    6 FORMAT('     ')
 5551 FORMAT(T1,'<<<End of the ELEMENTS data block',/,'     ')
 5552 FORMAT(T1,'<<<End of the CONNECTIONS data block',/,'     ')
! 
! 
! 
      IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
         WRITE(MESH_Unit, FMT = '(T1, A)') '>>>CONNECTIONS: Cartesian coordinates'
         WRITE(MESH_Unit, FMT = '(T1, A)') '&Units  length_units = "m" /'
      ELSE
         WRITE(MESH_Unit, FMT = 7)
      END IF 
! 
! 
    7 FORMAT('CONNE')
!
!-----ASSIGN HORIZONTAL CONNECTIONS.
!
!     NII=' '
!     NII1=' '
!     DO 8 I=NSKIP2,NELEMT
      DO 8 I8=NSKIP2-1,NELEMT-1
!        MI = MOD(I,100)
         MI = MOD(I8,100)
         I  = I8+1
!        II = I/100
!        IF(II.GE.1) NII=NA(II)
!        NII=NA(II+1)

         IF(MOD(I,100) == 0) THEN
            II = I/100
         ELSE
            II = I/100+1
         END IF
         NII = NA(II)

!        MI1 = MOD(I-1,100)
         MI1 = MOD(I8-1,100)
!        II1 = (I-1)/100
!        IF(II1.GE.1) NII1=NA(II1)
!        NII1=NA(II1+1)

         IF(MOD(I8,100) == 0) THEN
            II1 = (I8)/100
         ELSE
            II1 = (I8)/100+1
         END IF
         NII1 = NA(II1)
! 
! - GJM: Begin modification for 8-character elements
! 
         IF(ElemNameLength.EQ.8) THEN
            DO J=1,NLAY
               MJ  = MOD(J-1,36)+1
               NOV = NA((J-1)/36+11)
               AIJ = A(I-1)*H(J)
!
               IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
                  WRITE(MESH_Unit,5520) NOV,NA(MJ),NII1,MI1,   &
     &                                  NOV,NA(MJ),NII,MI, AIJ, D2(I-1),D1(I),1
               ELSE
                  WRITE(MESH_Unit,5009) 'ZZZ',NOV,NA(MJ),NII1,MI1,   &
     &                                  'ZZZ',NOV,NA(MJ),NII,MI,D2(I-1),D1(I),AIJ
               END IF
!
            END DO
         ELSE
            DO J=1,NLAY
               MJ  = MOD(J-1,36)+1
               NOV = NA((J-1)/36+11)
               AIJ = A(I-1)*H(J)
!
               IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
                  WRITE(MESH_Unit,5530) NOV,NA(MJ),NII1,MI1,   &
     &                                  NOV,NA(MJ),NII,MI, AIJ, D2(I-1),D1(I),1
               ELSE
                  WRITE(MESH_Unit,9) NOV,NA(MJ),NII1,MI1,   &
     &                               NOV,NA(MJ),NII,MI,D2(I-1),D1(I),AIJ
               END IF
!
            END DO
         END IF
 5530 FORMAT(3(a1),i2.2,3(a1),i2.2,'  &Conx  A=',1ES17.10,', d1=',1ES15.8,', d2=',1ES15.8,', dir=',i1,' /')
 5520 FORMAT('ZZZ',3(a1),i2.2,'ZZZ',3(a1),i2.2,'  &Conx  A=',1ES17.10,', d1=',1ES15.8,', d2=',1ES15.8,', dir=',i1,' /')
! 
 5009 FORMAT(A3,3A1,I2.2,A3,3A1,I2.2,13X,'1',3(1PE10.4))
    9 FORMAT(3A1,I2.2,3A1,I2.2,19X,'1',3(1PE10.4))
! 
! - GJM: End modification for 8-character elements
! 
    8 CONTINUE
!
!     MJ RUNS FROM 1 TO 36 FOR CONSECUTIVE LAYERS.
!     NOV RUNS THROUGH A, B, C, ... FOR THE 1ST, 2ND, 3RD, ... SET
!     OF 35 LAYERS.
!    
!
      IF(NLAY.LE.1) GO TO 42
!
!-----NOW FOR THE VERTICAL CONNECTIONS.
!
!     NII=' '
!     DO 40 I=NSKIP1,NELEMT
      DO 40 I8=NSKIP1-1,NELEMT-1
!        MI  = MOD(I,100)
         MI  = MOD(I8,100)
         I   = I8+1
!
         IF(MOD(I,100) == 0) THEN
            II = I/100
         ELSE
            II = I/100+1
         END IF
         NII = NA(II)
!
!        IF(II.GE.1) NII=NA(II)
         AIJ = V(I)
! 
! - GJM: Begin modification for 8-character elements
! 
         IF(ElemNameLength.EQ.8) THEN
            DO J=2,NLAY
               J1   = J-1
               MJ   = MOD(J-1,36)+1
               NOV  = NA((J-1)/36+11)
               MJ1  = MOD(J1-1,36)+1
               NOV1 = NA((J1-1)/36+11)
               D1U  = H(J1)/2.0d0
               D2U  = H(J)/2.0d0
!
               IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
                  WRITE(MESH_Unit,5521) NOV1,NA(MJ1),NII,MI,   &
     &                                  NOV,NA(MJ),NII,MI,AIJ,D1U,D2U,3,1.0d0
               ELSE
                  WRITE(MESH_Unit,5041) 'ZZZ',NOV1,NA(MJ1),NII,MI,   &
     &                                  'ZZZ',NOV,NA(MJ),NII,MI,D1U,D2U,AIJ
               END IF
!
            END DO
         ELSE
            DO J=2,NLAY
               J1   = J-1
               MJ   = MOD(J-1,36)+1
               NOV  = NA((J-1)/36+11)
               MJ1  = MOD(J1-1,36)+1
               NOV1 = NA((J1-1)/36+11)
               D1U  = H(J1)/2.0d0
               D2U  = H(J)/2.0d0
!
               IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
                  WRITE(MESH_Unit,5531) NOV1,NA(MJ1),NII,MI,   &
     &                                  NOV,NA(MJ),NII,MI,AIJ,D1U,D2U,3,1.0d0
               ELSE
                  WRITE(MESH_Unit,41) NOV1,NA(MJ1),NII,MI,   &
     &                                NOV,NA(MJ),NII,MI,D1U,D2U,AIJ
               END IF
!
            END DO
         END IF
! 
! 
 5531 FORMAT(3(a1),i2.2,3(a1),i2.2,'  &Conx  A=',1ES17.10,', d1=',1ES15.8,', d2=',1ES15.8,', dir=',i1,', beta=',1ES17.10,' /')
 5521 FORMAT('ZZZ',3(a1),i2.2,'ZZZ',3(a1),i2.2,'  &Conx  A=',1ES17.10,', d1=',1ES15.8,', d2=',1ES15.8, &
     &        ', dir=',i1,', beta=',1ES17.10,' /')
! 
 5041 FORMAT(A3,3A1,I2.2,A3,3A1,I2.2,13X,'3',3(1PE10.4),'1.')
   41 FORMAT(3A1,I2.2,3A1,I2.2,19X,'3',3(1PE10.4),' 1.0e0')
! 
! - GJM: End modification for 8-character elements
! 
   40 CONTINUE
!
   42 CONTINUE
!
!
!
      IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
         WRITE(MESH_Unit, FMT = 5555) 
       ELSE
         WRITE(MESH_Unit, FMT = 6)
      END IF 
!
 5555 FORMAT(T1,'<<<End of the CONNECTIONS data block',/,'     ')
!
!
!
      REWIND (UNIT = MESH_Unit)
      READ(MESH_Unit,43) DOM
!
      L=0
! 
! - GJM: Begin modification for 8-character elements
! 
      DO 44 I = NSKIP1,NELEMT
         DO 44 J = 1,NLAY
            L = L+1
            IF(ElemNameLength == 8) THEN
               READ(MESH_Unit,5043) name(L)
            ELSE
               READ(MESH_Unit,43) name(L)(1:5)
            END IF
   44 CONTINUE
!
 5043 FORMAT(A8)
   43 FORMAT(A5)
! 
! - GJM: End modification for 8-character elements
!
      CALL PRZ2D(K,NSKIP1,NELEMT,NLAY)
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of WRZ2D
!
!
      RETURN
!
      END SUBROUTINE WRZ2D
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE PRZ2D(KK,NSKIP1,NELEMT,NLAY)
! 
         USE MeshMaker_Data
! 
! 
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
! 
      SAVE ICALL
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of PRZ2D
!
!
! 
!     FIND LOCATION OF ELEMENT NAME WITH R-INDEX I, Z-INDEX K.
!     MESH ORGANIZED "BY COLUMNS" (KK = 1)
!
      NIK1(I,K)=(I-1)*NLAY+K
!
!     MESH ORGANIZED "BY LAYERS" (KK = 2)
!
      NIK2(I,K)=(K-1)*(NELEMT-NSKIP1+1)+I
!
      ICALL=ICALL+1
      IF(ICALL == 1) WRITE(*,899)
!
! 899 FORMAT('PRZ2D             1.0   27 MARCH     1991',6X,'MAKE STRUCTURED PRINTOUT OF 2-D R-Z MESH')
  899 FORMAT('PRZ2D             1.5   29 July      2014',6X,'MAKE STRUCTURED PRINTOUT OF 2-D R-Z MESH')
!
      NRZ = NELEMT*NLAY
      WRITE(*,31) NELEMT,NLAY,NRZ
   31 FORMAT(/,' ',131('*')/' *',20X,'2-D R-Z MESH WITH NR*NLAY = ',I4,' *',I4,'  =',I4,' GRID BLOCKS',52X,'*'/' ',131('*'))
!
      WRITE(*,18) NLAY,NELEMT
   18 FORMAT(' *',129X,'*'/   &
     &   ' *',20X,'THE MESH WILL BE PRINTED AS VERTICAL SLICES',66X,'*',/,' *',129X,'*',/,   &
     &   ' *',20X,'LAYERS GO FROM K = 1 TO K = NLAY =',I4,71X,'*',/,' *',129X,'*',/,  &
     &   ' *',20X,'RADIAL GRID BLOCKS GO IN COLUMNS FROM I = 1 TO I = NR =',I4,50X,'*',/,' *',129X,'*'/' ',131('*')/)
! 
! - GJM: Begin modification for 8-character elements
! 
      IF(ElemNameLength == 8) THEN
         WRITE(*,5004)
         DO K=1,NLAY
            IF(KK == 1) WRITE(*,5006) K,(name(NIK1(I,K)), I=NSKIP1,NELEMT)
            IF(KK == 2) WRITE(*,5006) K,(name(NIK2(I,K)), I=NSKIP1,NELEMT)
         END DO
      ELSE
         WRITE(*,4)
         DO K=1,NLAY
            IF(KK == 1) WRITE(*,6) K,(name(NIK1(I,K))(1:5), I=NSKIP1,NELEMT)
            IF(KK == 2) WRITE(*,6) K,(name(NIK2(I,K))(1:5), I=NSKIP1,NELEMT)
         END DO
      END IF
! 
 5004 FORMAT('   COLUMN I =    1        2        3        4        5        6        7        8        9       10       11       12',/,' LAYERS')
    4 FORMAT('   COLUMN I =  1     2     3     4     5     6     7     8     9    10    11    12    13    14    15    16    17    18    19    20',/,' LAYERS')
 5006 FORMAT('  K = ',I4,2X,12(1X,A8),/,80(12X,12(1X,A8),/))
    6 FORMAT('  K = ',I4,2X,20(1X,A5),/,50(12X,20(1X,A5),/))
! 
! - GJM: End modification for 8-character elements
! 
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of PRZ2D
!
!
      RETURN
!
      END SUBROUTINE PRZ2D
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE GXYZ
! 
         USE Grid_Gener_parameters
         USE Het_Region_Definition
         USE Boundary_Definition
! 
         USE MeshMaker_Data
! 
! 
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
! 
!
!-----PROGRAM FOR GENERATING 3-D X-Y-Z MESHES
!
!     GRID BLOCK DIMENSIONS ARE DX, DY, AND DZ
!
! 
! -------
! ... Character variables     
! -------
! 
      CHARACTER NXYZ*2,bound_id*1
      CHARACTER*5 DOM,ELNAME,ELNAM1,ONOMA,medium
! 
      DIMENSION DX(2500),DY(2500),DZ(2500)
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ Begin addition for optimal numbering to minimize bandwidth 
! ----------------
! 
      DIMENSION DLXYZ(2500,3),DDXYZ(3),RRR(3),NLXYZ(3)
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ End addition for optimal numbering to minimize bandwidth 
! ----------------
! 
      DIMENSION D1(2),D2(2),D3(2)
!
      SAVE ICALL,NX,NY,NZ
      DATA NX,NY,NZ/0,0,0/
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of GXYZ
!
!
      ICALL=ICALL+1
      IF(ICALL == 1) WRITE(*,6000)
!
! 899 FORMAT(/6X,'GXYZ     1.0      18 MARCH     1991',6X,
 6000 FORMAT(/,'GXYZ              2.0   18 September 2000',6X,'GENERATE 1, 2, OR 3-D CARTESIAN MESH')
!
      REWIND (UNIT = MESH_Unit)
!
!
!
      READ(*,*) DEG, X_ref, Y_ref, Z_ref
!
      X_ref = X_ref * factor
      Y_ref = Y_ref * factor 
      Z_ref = Z_ref * factor 
!
 5000 FORMAT(A2,3X,I5,E10.4)
 5001 FORMAT(8E10.4)
!
  500 READ(*,5000) NXYZ,NO,DEL
!
!
!
      SELECT CASE(NXYZ)
!
! -------
! ... X-increments
! -------
!
      CASE('NX')
!
         NX1 = NX + 1
         NX  = NX + NO
         IF(DEL /= 0.0d0) THEN
            DX(NX1:NX) = DEL
         ELSE
            READ(*,5001) (DX(I),I=NX1,NX)
         END IF
!
         DX(NX1:NX) = DX(NX1:NX) * factor
!
         GO TO 500
!
! -------
! ... Y-increments
! -------
!
      CASE('NY')
!
         NY1 = NY + 1
         NY  = NY + NO
         IF(DEL /= 0.0d0) THEN
            DY(NY1:NY) = DEL
         ELSE
            READ(*,5001) (DY(I),I=NY1,NY)
         END IF
!
         DY(NY1:NY) = DY(NY1:NY) * factor
!
         GO TO 500
!
! -------
! ... Z-increments
! -------
!
      CASE('NZ')
!
         NZ1 = NZ + 1
         NZ  = NZ + NO
         IF(DEL /= 0.0d0) THEN
            DZ(NZ1:NZ) = DEL
         ELSE
            READ(*,5001) (DZ(I),I=NZ1,NZ)
         END IF
!
         DZ(NZ1:NZ) = DZ(NZ1:NZ) * factor
!
         GO TO 500
!
! -------
! ... End of increment inputs
! -------
!
      CASE DEFAULT 
         CONTINUE
      END SELECT
!
!
!
      RAD  = PI*DEG/180.
      BET  = SIN(RAD)
      BETA = COS(RAD)
!
!-----MAKE PRINTOUT OF MESH SPECIFICATIONS.
!
      NN = NX*NY*NZ
!
      WRITE(*,40) NX,NY,NZ,NN,DEG
   40 FORMAT(/,' GENERATE CARTESIAN MESH WITH Nx*Ny*Nz = ',   &
     &I5,' *',I5,' *',I5,'  =  ',I6,'  GRID BLOCKS'/1X,131('*')//   &
     &'   THE X-AXIS IS HORIZONTAL; THE Y-AXIS IS ROTATED BY ',1pE11.4,' DEGREES AGAINST THE HORIZONTAL'//   &
     &'   THE GRID INCREMENTS ARE AS FOLLOWS')
!
      PRINT 41,NX,(NA(I),I=1,9),(DX(I),I=1,NX)
   41 FORMAT(//'   DELTA-X  (',I4,' INCREMENTS)'//13X,9(A1,10X),'10'//(10X,10(1X,1pE11.4)))
!
      PRINT 441,NX,(NA(I),I=1,9),(X_ref + SUM(DX(1:I)),I=1,NX)
  441 FORMAT(//'   X  (',I4,' COORDINATES)'//13X,9(A1,11X),'10'//(10X,10(1X,1pE11.4)))
!
      PRINT 42,NY,(NA(I),I=1,9),(DY(I),I=1,NY)
   42 FORMAT(//'   DELTA-Y  (',I4,' INCREMENTS)'//13X,9(A1,10X),'10'//(10X,10(1X,1pE11.4)))
!
      PRINT 442,NY,(NA(I),I=1,9),(Y_ref + SUM(DY(1:I)),I=1,NY)
  442 FORMAT(//'   Y  (',I4,' COORDINATES)'//13X,9(A1,11X),'10'//(10X,10(1X,1pE11.4)))
!
      PRINT 43,NZ,(NA(I),I=1,9),(DZ(I),I=1,NZ)
   43 FORMAT(//'   DELTA-Z  (',I4,' INCREMENTS)'//13X,9(A1,10X),'10'//(10X,10(1X,1pE11.4)))
!
      PRINT 443,NZ,(NA(I),I=1,9),(Z_ref - SUM(DZ(1:I)),I=1,NZ)
  443 FORMAT(//'   Z  (',I4,' COORDINATES)'//13X,9(A1,11X),'10'//(10X,10(1X,1pE11.4)))
!
      PRINT 444 
  444 FORMAT(//'   Z  LAYERS - Beginning from the top, moving downward:')
         PRINT 445, Z_ref-DZ(1), Z_ref, 1
      DO i = 2,NZ
         PRINT 445, Z_ref-SUM(DZ(1:i)), Z_ref-SUM(DZ(1:i-1)), i
      END DO
  445 FORMAT(T5,2(1X,1pE15.8),3x,'! Layer ',i3.3)
!
      PRINT 44
   44 FORMAT(/' ',131('*'))
!
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ Begin addition for optimal numbering to minimize bandwidth 
! ----------------
! 
!
      IF(NX >= NY .AND. NX >= NZ) THEN
!
         NLXYZ(3)      = NX
         DLXYZ(1:NX,3) = DX(1:NX)
!
         IF(NY >= NZ) THEN
!
            NLXYZ(2)      = NY
            DLXYZ(1:NY,2) = DY(1:NY)
!
            NLXYZ(1)      = NZ
            DLXYZ(1:NZ,1) = DZ(1:NZ)
!
            NNNUUU = 1
!
            GO TO 1000
!
         ELSE
!
            NLXYZ(2)      = NZ
            DLXYZ(1:NZ,2) = DZ(1:NZ)
!
            NLXYZ(1)      = NY
            DLXYZ(1:NY,1) = DY(1:NY)
!
            NNNUUU = 2
!
            GO TO 1000
!
         END IF
!
      END IF
!
!
!
      IF(NY >= NX .AND. NY >= NZ) THEN
!
         NLXYZ(3)      = NY
         DLXYZ(1:NY,3) = DY(1:NY)
!
         IF(NX >= NZ) THEN
!
            NLXYZ(2)      = NX
            DLXYZ(1:NX,2) = DX(1:NX)
!
            NLXYZ(1)      = NZ
            DLXYZ(1:NZ,1) = DZ(1:NZ)
!
            NNNUUU = 3
!
            GO TO 1000
!
         ELSE
!
            NLXYZ(2)      = NZ
            DLXYZ(1:NZ,2) = DZ(1:NZ)
!
            NLXYZ(1)      = NX
            DLXYZ(1:NX,1) = DX(1:NX)
!
            NNNUUU = 4
!
            GO TO 1000
!
         END IF
!
      END IF
!
!
!
      IF(NZ >= NX .AND. NZ >= NY) THEN
!
         NLXYZ(3)      = NZ
         DLXYZ(1:NZ,3) = DZ(1:NZ)
!
         IF(NX >= NY) THEN
!
            NLXYZ(2)      = NX
            DLXYZ(1:NX,2) = DX(1:NX)
!
            NLXYZ(1)      = NY
            DLXYZ(1:NY,1) = DY(1:NY)
!
            NNNUUU = 5
!
            GO TO 1000
!
         ELSE
!
            NLXYZ(2)      = NY
            DLXYZ(1:NY,2) = DY(1:NY)
!
            NLXYZ(1)      = NX
            DLXYZ(1:NX,1) = DX(1:NX)
!
            NNNUUU = 6
!
         END IF
!
      END IF
!
!
!
 1000 IF((NNNUUU == 1) .OR. (NNNUUU == 3)) THEN
         RRR(1) =-1.0d0
      ELSE
         RRR(1) = 1.0d0
      END IF
!
      IF((NNNUUU == 2) .OR. (NNNUUU == 4)) THEN
         RRR(2) =-1.0d0
      ELSE
         RRR(2) = 1.0d0
      END IF
!
      IF((NNNUUU == 5) .OR. (NNNUUU == 6)) THEN 
         RRR(3) =-1.0d0
      ELSE
         RRR(3) = 1.0d0
      END IF
!
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ End addition for optimal numbering to minimize bandwidth 
! ----------------
! 
!
      ILOOP = 0
   30 CONTINUE
      ILOOP = ILOOP+1
!
! 
! - GJM: Begin modification for 8-character elements
! 
      IF(ILOOP == 1) THEN 
         IF_Format0: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
            WRITE(MESH_Unit, FMT = '(T1, A)') '>>>ELEMENTS: Cartesian coordinates'
         WRITE(MESH_Unit, FMT = '(T1, A)') '&Units  length_units = "m" /'
         ELSE
! 
            IF(ElemNameLength == 8) THEN
               IF(ILOOP == 1) WRITE(MESH_Unit,5020)
            ELSE
               IF(ILOOP == 1) WRITE(MESH_Unit,20)
            END IF
! 
         END IF IF_Format0
      END IF

! 
      IF(ILOOP == 2) THEN 
         IF_Format1: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
            WRITE(MESH_Unit, FMT = '(T1, A)') '>>>CONNECTIONS: Cartesian coordinates'
         WRITE(MESH_Unit, FMT = '(T1, A)') '&Units  length_units = "m" /'
         ELSE
            WRITE(MESH_Unit,21)
         END IF IF_Format1
      END IF
! 
 5020 FORMAT('ELEMEext2')
   20 FORMAT('ELEME')
   21 FORMAT('CONNE')
! 
! - GJM: End modification for 8-character elements
! 
!
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ Begin addition  
! ........   (a) for optimal numbering to minimize bandwidth 
! ........   (b) to introduce diffenet numbering/naming system to allow
! ........       the use of the MINC facility for large grids
! ----------------
! 
!
      XI = RRR(3)*DLXYZ(1,3)/2.0d0
!
      DO 1 I=1,NLXYZ(3)
! 
         I1    = I+1
         D1(1) = DLXYZ(I,3)/2.0d0
         IF(I < NLXYZ(3)) D1(2) = DLXYZ(I1,3)/2.0d0
         IF(I > 1) XI = XI + RRR(3)*DLXYZ(I,3)/2.0d0 + RRR(3)*DLXYZ(I-1,3)/2.0d0
! 
         YJ = RRR(2)*DLXYZ(1,2)/2.0d0
! 
! 
         DO 1 J=1,NLXYZ(2)
! 
            J1    = J+1
            D2(1) = DLXYZ(J,2)/2.0d0
            IF(J < NLXYZ(2)) D2(2) = DLXYZ(J1,2)/2.0d0
            IF(J > 1) YJ = YJ + RRR(2)*DLXYZ(J,2)/2.0d0 + RRR(2)*DLXYZ(J-1,2)/2.0d0
! 
            ZK = RRR(1)*DLXYZ(1,1)/2.0d0
! 
! 
            DO 1 K=1,NLXYZ(1)
! 
               K1    = K+1
               D3(1) = DLXYZ(K,1)/2.0d0
               IF(K < NLXYZ(1)) D3(2) = DLXYZ(K1,1)/2.0d0
               IF(K > 1) ZK = ZK + RRR(1)*DLXYZ(K,1)/2.0d0 + RRR(1)*DLXYZ(K-1,1)/2.0d0
! 
               IF(ILOOP == 2) GOTO 31
! 
               V = DLXYZ(I,3)*DLXYZ(J,2)*DLXYZ(K,1)
! 
               AIJ = 0.0d0
! 
               IF((NNNUUU == 1) .OR. (NNNUUU == 3)) THEN
                 IF(K == 1)  AIJ = AIJ + DLXYZ(J,2)*DLXYZ(I,3)
                 IF(K == NZ) AIJ = AIJ + DLXYZ(J,2)*DLXYZ(I,3)
               END IF
! 
               IF((NNNUUU == 2) .OR. (NNNUUU == 4)) THEN
                 IF(J == 1)  AIJ = AIJ + DLXYZ(K,1)*DLXYZ(I,3)
                 IF(J == NZ) AIJ = AIJ + DLXYZ(K,1)*DLXYZ(I,3)
               END IF
! 
               IF((NNNUUU == 5) .OR. (NNNUUU == 6)) THEN
                 IF(I == 1)  AIJ = AIJ + DLXYZ(K,1)*DLXYZ(J,2)
                 IF(I == NZ) AIJ = AIJ + DLXYZ(K,1)*DLXYZ(J,2)
               END IF
! 
! 
! 
               SELECT CASE(NNNUUU)
               CASE(1)
                  DDXYZ(1) = XI+X_ref
                  DDXYZ(2) = YJ+Y_ref
                  DDXYZ(3) = ZK+Z_ref
               CASE(2)
                  DDXYZ(1) = XI+X_ref
                  DDXYZ(2) = ZK+Z_ref
                  DDXYZ(3) = YJ+Y_ref
               CASE(3)
                  DDXYZ(1) = YJ+Y_ref
                  DDXYZ(2) = XI+X_ref
                  DDXYZ(3) = ZK+Z_ref
               CASE(4)
                  DDXYZ(1) = ZK+Z_ref
                  DDXYZ(2) = XI+X_ref
                  DDXYZ(3) = YJ+Y_ref
               CASE(5)
                  DDXYZ(1) = YJ+Y_ref
                  DDXYZ(2) = ZK+Z_ref
                  DDXYZ(3) = XI+X_ref
               CASE(6)
                  DDXYZ(1) = ZK+Z_ref
                  DDXYZ(2) = YJ+Y_ref
                  DDXYZ(3) = XI+X_ref
               END SELECT
! 
! 
               ELNAME = ONOMA(NLXYZ(1),NLXYZ(2),NLXYZ(3),i,j,k)
! 
! - GJM: Begin modification for 8-character elements
!
! ----------------
! ............ When the domain is homogeneous, assign the standard domain number/name
! ----------------
! 
               IF_Media: IF(Flag_HetRegions .EQV. .FALSE.) THEN
! 
                  DOM = '    1'
!
! ----------------
! ............ When the domain is heterogeneous ...
! ----------------
! 
               ELSE 
! 
! ............... Assign the dominant domain number/name
! 
                  IF_HetRegions: IF(Num_HetRegions == 1) THEN
! 
                     DOM = dominant_medium
! 
! ............... Assign the appropriate heterogeneous domain number/name
! 
                  ELSE
                     DOM = Media_Region(Flag_HetRegions,'Cartesian  ',dominant_medium,DDXYZ(1),DDXYZ(2),DDXYZ(3), media_by_number)
                  END IF IF_HetRegions
! 
               END IF IF_Media 
!
! ----------------
! ............ Determine whether this is a boundary cell
! ----------------
! 
               IF(Flag_Boundaries .AND. Num_Boundaries > 0) THEN 
! 
                  CALL Boundary_Status('Cartesian  ',DDXYZ(1),DDXYZ(2),DDXYZ(3),bound_id,medium,media_by_number)
! 
                  IF(bound_id == 'I' .OR. bound_id == 'V') THEN
                     DOM = medium
                  END IF
! 
               ELSE
                  bound_id = ' '
               END IF
! 
! 
 5400 FORMAT('ZZZ',A,'  &Elem  V=',1ES17.10,', MedNum=',A5,', x=',1ES17.10,', y=',1ES17.10,', z=',1ES17.10,' /')
 5401 FORMAT('ZZZ',A,'  &Elem  V=',1ES17.10,', MedNum=',A5,', x=',1ES17.10,', y=',1ES17.10,', z=',1ES17.10,', act="',A,'" /')
! 
 5500 FORMAT('ZZZ',A,'  &Elem  V=',1ES17.10,', MedNam="',A5,'", x=',1ES17.10,', y=',1ES17.10,', z=',1ES17.10,' /')
 5501 FORMAT('ZZZ',A,'  &Elem  V=',1ES17.10,', MedNam="',A5,'", x=',1ES17.10,', y=',1ES17.10,', z=',1ES17.10,', act="',A,'" /')
! 
 5402 FORMAT(A,'  &Elem  V=',1ES17.10,', MedNum=',A5,', x=',1ES17.10,', y=',1ES17.10,', z=',1ES17.10,' /')
 5403 FORMAT(A,'  &Elem  V=',1ES17.10,', MedNum=',A5,', x=',1ES17.10,', y=',1ES17.10,', z=',1ES17.10,', act="',A,'" /')
! 
 5502 FORMAT(A,'  &Elem  V=',1ES17.10,', MedNam="',A5,'", x=',1ES17.10,', y=',1ES17.10,', z=',1ES17.10,' /')
 5503 FORMAT(A,'  &Elem  V=',1ES17.10,', MedNam="',A5,'", x=',1ES17.10,', y=',1ES17.10,', z=',1ES17.10,', act="',A,'" /')
! 
! 
               IF_Format2: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
! 
                  IF(bound_id == ' ') THEN
                     IF(ElemNameLength.EQ.8) THEN
                        IF(DOM == '    1') THEN
                           WRITE(MESH_Unit, FMT = 5400) ELNAME,V,DOM, DDXYZ(1), DDXYZ(2), DDXYZ(3)
                        ELSE
                           WRITE(MESH_Unit, FMT = 5500) ELNAME,V,DOM, DDXYZ(1), DDXYZ(2), DDXYZ(3)
                        END IF
                     ELSE
                        IF(DOM == '    1') THEN
                           WRITE(MESH_Unit, FMT = 5402) ELNAME,V,DOM, DDXYZ(1), DDXYZ(2), DDXYZ(3)
                        ELSE
                           WRITE(MESH_Unit, FMT = 5502) ELNAME,V,DOM, DDXYZ(1), DDXYZ(2), DDXYZ(3)
                        END IF
                     END IF
                  ELSE
                     IF(ElemNameLength.EQ.8) THEN
                        IF(DOM == '    1') THEN
                           WRITE(MESH_Unit, FMT = 5401) ELNAME,V,DOM, DDXYZ(1), DDXYZ(2), DDXYZ(3),bound_id
                        ELSE
                           WRITE(MESH_Unit, FMT = 5501) ELNAME,V,DOM, DDXYZ(1), DDXYZ(2), DDXYZ(3),bound_id
                        END IF
                     ELSE
                        IF(DOM == '    1') THEN
                           WRITE(MESH_Unit, FMT = 5403) ELNAME,V,DOM, DDXYZ(1), DDXYZ(2), DDXYZ(3),bound_id
                        ELSE
                           WRITE(MESH_Unit, FMT = 5503) ELNAME,V,DOM, DDXYZ(1), DDXYZ(2), DDXYZ(3),bound_id
                        END IF
                     END IF
                  END IF
! 
               ELSE
! 
                  IF(ElemNameLength == 8) THEN
                     WRITE(MESH_Unit,5010) 'ZZZ',ELNAME,DOM,V,AIJ, DDXYZ(1), DDXYZ(2), DDXYZ(3),bound_id
                  ELSE
                     WRITE(MESH_Unit,10) ELNAME,DOM,V,AIJ, DDXYZ(1), DDXYZ(2), DDXYZ(3),bound_id
                  END IF
! 
               END IF IF_Format2



! 
! - GJM: End modification for 8-character elements
! 
               GO TO 1

   31          CONTINUE
! 
               A12 = DLXYZ(J,2)*DLXYZ(K,1)
               A13 = DLXYZ(I,3)*DLXYZ(K,1)
               A23 = DLXYZ(I,3)*DLXYZ(J,2)
! 
               ELNAME = ONOMA(NLXYZ(1),NLXYZ(2),NLXYZ(3),i,j,k)
! 
 5530 FORMAT(A,A,'  &ConX  A=',1ES17.10,', d1=',1ES15.8,', d2=',1ES15.8,', dir=',A1,' /')
 5531 FORMAT(A,A,'  &Conx  A=',1ES17.10,', d1=',1ES15.8,', d2=',1ES15.8,', dir=',A1,', beta=',1ES15.8,' /')
! 
 5520 FORMAT('ZZZ',A,'ZZZ',A,'  &ConX  A=',1ES17.10,', d1=',1ES15.8,', d2=',1ES15.8,', dir=',A1,' /')
 5521 FORMAT('ZZZ',A,'ZZZ',A,'  &Conx  A=',1ES17.10,', d1=',1ES15.8,', d2=',1ES15.8,', dir=',A1,', beta=',1ES15.8,' /')
! 
               IF(I < NLXYZ(3)) THEN 
! 
                  ELNAM1 = ONOMA(NLXYZ(1),NLXYZ(2),NLXYZ(3),i+1,j,k)
! 
! - GJM: Begin modification for 8-character elements
! 
               IF_Format3: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
                  IF(ElemNameLength == 8) THEN
                     IF((NNNUUU == 1) .OR. (NNNUUU == 2)) WRITE(MESH_Unit,5520) ELNAME,ELNAM1,A12,D1(1),D1(2),NA(2)
                     IF((NNNUUU == 3) .OR. (NNNUUU == 4)) WRITE(MESH_Unit,5521) ELNAME,ELNAM1,A12,D1(1),D1(2),NA(3),BET
                     IF((NNNUUU == 5) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,5521) ELNAME,ELNAM1,A12,D1(1),D1(2),NA(4),BETA
                  ELSE
                     IF((NNNUUU == 1) .OR. (NNNUUU == 2)) WRITE(MESH_Unit,5530) ELNAME,ELNAM1,A12,D1(1),D1(2),NA(2)
                     IF((NNNUUU == 3) .OR. (NNNUUU == 4)) WRITE(MESH_Unit,5531) ELNAME,ELNAM1,A12,D1(1),D1(2),NA(3),BET
                     IF((NNNUUU == 5) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,5531) ELNAME,ELNAM1,A12,D1(1),D1(2),NA(4),BETA
                  END IF
               ELSE
                  IF(ElemNameLength == 8) THEN
                     IF((NNNUUU == 1) .OR. (NNNUUU == 2)) WRITE(MESH_Unit,5011) 'ZZZ',ELNAME,'ZZZ',ELNAM1,NA(2),D1(1),D1(2),A12
                     IF((NNNUUU == 3) .OR. (NNNUUU == 4)) WRITE(MESH_Unit,5011) 'ZZZ',ELNAME,'ZZZ',ELNAM1,NA(3),D1(1),D1(2),A12,BET
                     IF((NNNUUU == 5) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,5011) 'ZZZ',ELNAME,'ZZZ',ELNAM1,NA(4),D1(1),D1(2),A12,BETA
                  ELSE
                     IF((NNNUUU == 1) .OR. (NNNUUU == 2)) WRITE(MESH_Unit,11) ELNAME,ELNAM1,NA(2),D1(1),D1(2),A12
                     IF((NNNUUU == 3) .OR. (NNNUUU == 4)) WRITE(MESH_Unit,11) ELNAME,ELNAM1,NA(3),D1(1),D1(2),A12,BET
                     IF((NNNUUU == 5) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,11) ELNAME,ELNAM1,NA(4),D1(1),D1(2),A12,BETA
                  END IF
               END IF IF_Format3
! 
! - GJM: End modification for 8-character elements
! 
               END IF
! 
! 
! 
               IF(J < NLXYZ(2)) THEN 
! 
                  ELNAM1 = ONOMA(NLXYZ(1),NLXYZ(2),NLXYZ(3),i,j+1,k)
! 
! 
! - GJM: Begin modification for 8-character elements
! 
               IF_Format4: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
                  IF(ElemNameLength == 8) THEN
                     IF((NNNUUU == 3) .OR. (NNNUUU == 5)) WRITE(MESH_Unit,5520) ELNAME,ELNAM1,A13,D2(1),D2(2),NA(2)
                     IF((NNNUUU == 1) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,5521) ELNAME,ELNAM1,A13,D2(1),D2(2),NA(3),BET
                     IF((NNNUUU == 2) .OR. (NNNUUU == 4)) WRITE(MESH_Unit,5521) ELNAME,ELNAM1,A13,D2(1),D2(2),NA(4),BETA
                  ELSE
                     IF((NNNUUU == 3) .OR. (NNNUUU == 5)) WRITE(MESH_Unit,5530) ELNAME,ELNAM1,A13,D2(1),D2(2),NA(2)
                     IF((NNNUUU == 1) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,5531) ELNAME,ELNAM1,A13,D2(1),D2(2),NA(3),BET
                     IF((NNNUUU == 2) .OR. (NNNUUU == 4)) WRITE(MESH_Unit,5531) ELNAME,ELNAM1,A13,D2(1),D2(2),NA(4),BETA
                  END IF
               ELSE
                  IF(ElemNameLength == 8) THEN
                     IF((NNNUUU == 3) .OR. (NNNUUU == 5)) WRITE(MESH_Unit,5011) 'ZZZ',ELNAME,'ZZZ',ELNAM1,NA(2),D2(1),D2(2),A13
                     IF((NNNUUU == 1) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,5011) 'ZZZ',ELNAME,'ZZZ',ELNAM1,NA(3),D2(1),D2(2),A13,BET
                     IF((NNNUUU == 2) .OR. (NNNUUU == 4)) WRITE(MESH_Unit,5011) 'ZZZ',ELNAME,'ZZZ',ELNAM1,NA(4),D2(1),D2(2),A13,BETA
                  ELSE
                     IF((NNNUUU == 3) .OR. (NNNUUU == 5)) WRITE(MESH_Unit,11) ELNAME,ELNAM1,NA(2),D2(1),D2(2),A13
                     IF((NNNUUU == 1) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,11) ELNAME,ELNAM1,NA(3),D2(1),D2(2),A13,BET
                     IF((NNNUUU == 2) .OR. (NNNUUU == 4)) WRITE(MESH_Unit,11) ELNAME,ELNAM1,NA(4),D2(1),D2(2),A13,BETA
                  END IF
               END IF IF_Format4
! 
! - GJM: End modification for 8-character elements
! 
               END IF
! 
! 
!
               IF(K < NLXYZ(1)) THEN 
! 
                  ELNAM1 = ONOMA(NLXYZ(1),NLXYZ(2),NLXYZ(3),i,j,k+1)
! 
! - GJM: Begin modification for 8-character elements
! 
               IF_Format5: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
                  IF(ElemNameLength == 8) THEN
                     IF((NNNUUU == 4) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,5520) ELNAME,ELNAM1,A23,D3(1),D3(2),NA(2)
                     IF((NNNUUU == 2) .OR. (NNNUUU == 5)) WRITE(MESH_Unit,5521) ELNAME,ELNAM1,A23,D3(1),D3(2),NA(3),BET
                     IF((NNNUUU == 1) .OR. (NNNUUU == 3)) WRITE(MESH_Unit,5521) ELNAME,ELNAM1,A23,D3(1),D3(2),NA(4),BETA
                  ELSE
                     IF((NNNUUU == 4) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,5530) ELNAME,ELNAM1,A23,D3(1),D3(2),NA(2)
                     IF((NNNUUU == 2) .OR. (NNNUUU == 5)) WRITE(MESH_Unit,5531) ELNAME,ELNAM1,A23,D3(1),D3(2),NA(3),BET
                     IF((NNNUUU == 1) .OR. (NNNUUU == 3)) WRITE(MESH_Unit,5531) ELNAME,ELNAM1,A23,D3(1),D3(2),NA(4),BETA
                  END IF
               ELSE
                  IF(ElemNameLength == 8) THEN
                     IF((NNNUUU == 4) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,5011) 'ZZZ',ELNAME,'ZZZ',ELNAM1,NA(2),D3(1),D3(2),A23
                     IF((NNNUUU == 2) .OR. (NNNUUU == 5)) WRITE(MESH_Unit,5011) 'ZZZ',ELNAME,'ZZZ',ELNAM1,NA(3),D3(1),D3(2),A23,BET
                     IF((NNNUUU == 1) .OR. (NNNUUU == 3)) WRITE(MESH_Unit,5011) 'ZZZ',ELNAME,'ZZZ',ELNAM1,NA(4),D3(1),D3(2),A23,BETA
                  ELSE
                     IF((NNNUUU == 4) .OR. (NNNUUU == 6)) WRITE(MESH_Unit,11) ELNAME,ELNAM1,NA(2),D3(1),D3(2),A23
                     IF((NNNUUU == 2) .OR. (NNNUUU == 5)) WRITE(MESH_Unit,11) ELNAME,ELNAM1,NA(3),D3(1),D3(2),A23,BET
                     IF((NNNUUU == 1) .OR. (NNNUUU == 3)) WRITE(MESH_Unit,11) ELNAME,ELNAM1,NA(4),D3(1),D3(2),A23,BETA
                  END IF
               END IF IF_Format5
! 
! - GJM: End modification for 8-character elements
! 
               END IF
!
    1 CONTINUE
 5010 FORMAT(A3,A5,7X,A5,2E10.4,10X,3F10.3,1X,A1)
   10 FORMAT(A5,10X,A5,2E10.4,10X,3E10.4,1X,A1)
 5011 FORMAT(2(A3,A5),13X,A1,4E10.4)
   11 FORMAT(2(A5),19X,A1,4E10.4)
!
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ End addition  
! ........   (a) for optimal numbering to minimize bandwidth 
! ........   (b) to introduce diffenet numbering/naming system to allow
! ........       the use of the MINC facility for large grids
! ----------------
! 
!
! 
      IF_Format6: IF(FormatType == 'New' .OR. FormatType == 'new' .OR. FormatType == 'NEW') THEN
         IF(ILOOP == 1) WRITE(MESH_Unit, FMT = 5551) 
         IF(ILOOP == 2) WRITE(MESH_Unit, FMT = 5552) 
      ELSE
         WRITE(MESH_Unit,FMT = 22)
      END IF IF_Format6
! 
 5551 FORMAT(T1,'<<<END of ELEMENTS data block',/,'     ')
 5552 FORMAT(T1,'<<<END of CONNECTIONS data block',/,'     ')
  22  FORMAT('     ')
! 
!
! 
      IF(ILOOP.LT.2) GOTO 30
!
      REWIND (UNIT = MESH_Unit)
      READ(MESH_Unit,50) DOM
! 
      L=0
! 
! - GJM: Begin modification for 8-character elements
! 
      IF(ElemNameLength == 8) THEN
         DO 56 I=1,NX
            DO 56 J=1,NY
               DO 56 K=1,NZ
                  L = L+1
                  READ(MESH_Unit,5050) name(L)
   56    CONTINUE
      ELSE
         DO 51 I=1,NX
            DO 51 J=1,NY
               DO 51 K=1,NZ
                  L = L+1
                  READ(MESH_Unit,50) name(L)(1:5)
   51    CONTINUE
      END IF
! 
 5050 FORMAT(A8)
   50 FORMAT(A5)
! 
! - GJM: End modification for 8-character elements
! 
      CALL PCAR(NLXYZ(3),NLXYZ(2),NLXYZ(1),NX,NY,NZ,NNNUUU)
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of GXYZ
!
!
      RETURN
      END SUBROUTINE GXYZ
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE PCAR(NX,NY,NZ,NXA,NYA,NZA,NNNUUU)
! 
         USE MeshMaker_Data
! 
! 
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
! 
!
!-----MAKE STRUCTURED PRINTOUT OF X-Y-Z GRID.
!
      CHARACTER*1 L1,L2,L3
      CHARACTER*2 M1,M2,M3
! 
      DIMENSION L1(6),L2(6),L3(6),M1(6),M2(6),M3(6)
!
      SAVE ICALL,L1,L2,L3,M1,M2,M3
!
      DATA ICALL/0/
      DATA L1/'K','J','K','I','J','I'/
      DATA L2/'J','K','I','K','I','J'/
      DATA L3/'I','I','J','J','K','K'/
      DATA M1/'NZ','NY','NZ','NX','NY','NX'/
      DATA M2/'NY','NZ','NX','NZ','NX','NY'/
      DATA M3/'NX','NX','NY','NY','NZ','NZ'/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of PCAR
!
!

!
!     FIND LOCATION OF ELEMENT NAME WITH X-INDEX I, Y-INDEX J,
!     AND Z-INDEX K
!
      NLOC(I,J,K)=(I-1)*NYA*NZA+(J-1)*NZA+K
!
      ICALL=ICALL+1
      IF(ICALL == 1) WRITE(*,899)
!
! 899 FORMAT(6X,'PCAR     1.0      25 MARCH     1991',6X,
! 899 FORMAT('PCAR              1.1   19 September 2000',6X,'MAKE STRUCTURED PRINTOUT OF CARTESIAN MESH')
  899 FORMAT('PCAR              1.5    2 August    2014',6X,'MAKE STRUCTURED PRINTOUT OF CARTESIAN MESH')
!
!
!
      PRINT 31, NXA,NYA,NZA
   31 format(/,' ',131('*'),   &
     &       /,' *',20X,'CARTESIAN MESH WITH NX*NY*NZ = ',I4,' *',I4,' *',I4,'  GRID BLOCKS',49X,'*',   &
     &       /,' ',131('*'))
!
      N1 = NZ
      N2 = NY
      N3 = NX
!
      PRINT 18,L1(NNNUUU),L1(NNNUUU),M1(NNNUUU),N1
   18 FORMAT(' *',129X,'*',/,' *',20X,'THE MESH WILL BE PRINTED AS SLICES FOR ',A1,' = 1 TO ',A1,' = ',A2,' =',I4,49X,'*')
!
      PRINT 19,L2(NNNUUU),L2(NNNUUU),M2(NNNUUU),N2
   19 FORMAT(' *',129X,'*',/,' *',20X,'IN EACH MESH SLICE, ROWS WILL GO FROM  ',A1,' = 1 TO ',A1,' = ',A2,' =',I4,49X,'*')
!
      PRINT 20,L3(NNNUUU),L3(NNNUUU),M3(NNNUUU),N3
   20 FORMAT(' *',129X,'*',/,' *',20X,'IN EACH ROW, COLUMNS WILL GO FROM      ',A1,' = 1 TO ',A1,' = ',A2,' =',I4,49X,'*',/,' *',129X,'*')
!
! 
! - GJM: Begin modification for 8-character elements
!    MBK: Modifications begin here (Oct 4, 2004) 
! 
      IF(ElemNameLength.EQ.8) THEN
         DO 102 K=1,N1
!            
            PRINT 3,L1(NNNUUU),K
            PRINT 5004,L3(NNNUUU)
            DO 105 J=1,N2
! 
               GOTO(121,122,123,124,125,126),NNNUUU
! 
  121          PRINT 5006,L2(NNNUUU),J,(name(NLOC(I,J,K)),I=1,N3)
               GOTO 105
! 
  122          PRINT 5006,L2(NNNUUU),J,(name(NLOC(I,K,J)),I=1,N3)
               GOTO 105
! 
  123          PRINT 5006,L2(NNNUUU),J,(name(NLOC(J,I,K)),I=1,N3)
               GOTO 105
! 
  124          PRINT 5006,L2(NNNUUU),J,(name(NLOC(K,I,J)),I=1,N3)
               GOTO 105
! 
  125          PRINT 5006,L2(NNNUUU),J,(name(NLOC(J,K,I)),I=1,N3)
               GOTO 105
! 
  126          PRINT 5006,L2(NNNUUU),J,(name(NLOC(K,J,I)),I=1,N3)
               GOTO 105
! 
  105       CONTINUE
  102    CONTINUE
      ELSE
! 
         DO 2 K=1,N1
            PRINT 3, L1(NNNUUU),K
            PRINT 4, L3(NNNUUU)
            DO 5 J=1,N2
! 
               GOTO(21,22,23,24,25,26),NNNUUU
! 
   21          PRINT 6,L2(NNNUUU),J,(name(NLOC(I,J,K))(1:5),I=1,N3)
               GOTO 5
! 
   22          PRINT 6,L2(NNNUUU),J,(name(NLOC(I,K,J))(1:5),I=1,N3)
               GOTO 5
! 
   23          PRINT 6,L2(NNNUUU),J,(name(NLOC(J,I,K))(1:5),I=1,N3)
               GOTO 5
! 
   24          PRINT 6,L2(NNNUUU),J,(name(NLOC(K,I,J))(1:5),I=1,N3)
               GOTO 5
! 
   25          PRINT 6,L2(NNNUUU),J,(name(NLOC(J,K,I))(1:5),I=1,N3)
               GOTO 5
! 
   26          PRINT 6,L2(NNNUUU),J,(name(NLOC(K,J,I))(1:5),I=1,N3)
               GOTO 5
! 
    5       CONTINUE
    2    CONTINUE
      END IF
!
!    MBK: Modifications end here (Oct 4, 2004)
! 
    3 FORMAT(' ',131('*')//' SLICE WITH ',A1,' =',I4/)
! 
 5004 FORMAT('   COLUMN ',A1,' =    1        2        3        4        5        6        7        8        9       10       11       12',/,' LAYERS')
    4 FORMAT('   COLUMN ',A1,' =  1     2     3     4     5     6     7     8     9    10    11    12    13    14    15    16    17    18    19    20',/,' ROWS')
 5006 FORMAT('  ',A1,' = ',I4,2X,12(1X,A8),/,80(12X,12(1X,A8),/) )
    6 FORMAT('  ',A1,' = ',I4,2X,20(1X,A5),/,50(12X,20(1X,A5),/) )
! 
! - GJM: End modification for 8-character elements
! 
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of PCAR
!
!
      RETURN
      END SUBROUTINE PCAR
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE MINC
!
         USE MeshMaker_Data
!
!-----THIS IS A MODIFIED AND ENHANCED VERSION OF PROGRAM GMINC ---
!
!-----THIS VERSION CAN ASSIGN VERTICAL CONNECTIONS BETWEEN MATRIX
!     CONTINUA (11-19-84).
!     FURTHER MODIFICATIONS (5-26-88):
!     (1) INCLUDE ELEMENT X,Y,Z-COORDINATES
!     (2) ELIMINATE "BOUNDARY ELEMENTS" FROM MINC-PROCEDURE
!
!***** GMINC WAS DEVELOPED BY KARSTEN PRUESS
!                          AT LAWRENCE BERKELEY LABORATORY. ************
!
!     THE PROGRAM GENERATES ONE-, TWO-, OR THREE-DIMENSIONAL MESHES
!     FOR FLOW SIMULATIONS IN FRACTURED POROUS MEDIA.
!
!     GMINC IMPLEMENTS THE METHOD OF
!                  MULTIPLE INTERACTING CONTINUA (MINC)
!     AS DEVELOPED BY PRUESS AND NARASIMHAN.
!
!     REFERENCES:
!
!     (1) K. PRUESS AND T.N. NARASIMHAN, A PRACTICAL METHOD FOR
!         MODELING FLUID AND HEAT FLOW IN FRACTURED POROUS MEDIA,
!         PAPER SPE-10509, PRESENTED AT THE SIXTH SPE-SYMPOSIUM ON
!         RESERVOIR SIMULATION, NEW ORLEANS, LA; (FEBRUARY 1982);
!         ALSO: SOCIETY OF PETROLEUM ENGINEERS JOURNAL, VOL. 25, NO.1,
!               PP. 14-26, 1985.
!
!     (2) K. PRUESS AND T.N. NARASIMHAN, ON FLUID RESERVES AND THE
!         PRODUCTION OF SUPERHEATED STEAM FROM FRACTURED, VAPOR-
!         DOMINATED GEOTHERMAL RESERVOIRS, J. GEOPHYS. RES. 87 (B11),
!         9329-9339, 1982.
!
!     (3) K. PRUESS AND K. KARASAKI, PROXIMITY FUNCTIONS FOR MODELING
!         FLUID AND HEAT FLOW IN RESERVOIRS WITH STOCHASTIC FRACTURE
!         DISTRIBUTIONS, PAPER PRESENTED AT EIGTH STANFORD WORKSHOP
!         ON GEOTHERMAL RESERVOIR ENGINEERING, STANFORD, CA.
!         (DECEMBER 1982).
!
!     (4) K. PRUESS, GMINC - A MESH GENERATOR FOR FLOW SIMULATIONS IN
!         FRACTURED RESERVOIRS, LAWRENCE BERKELEY LABORATORY REPORT
!         LBL-15227, 1983.
!
!***********************************************************************
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
!
      SAVE ICALL
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of MINC
!
!
      ICALL=ICALL+1
      IF(ICALL.EQ.1) WRITE(*,899)
  899 FORMAT(/,'MINC              1.0   22 JANUARY   1990',6X,   &
     &         'EXECUTIVE ROUTINE FOR MAKING A "SECONDARY" FRACTURED-POROUS MEDIUM MESH')
!
!-----READ DATA ON MINC-PARTITIONING.
!
      CALL PART
!
!!!!! CALL TO GEOM INSERTED 11-19-84, TO OBTAIN ALL VOL(M) BEFORE
!     PROCESSING CONNECTIONS.
      CALL GEOM
!
!-----READ ELEMENT DATA FROM FILE *MESH* AND PROCESS SEQUENTIALLY.
      CALL MINCME
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of MINC
!
!
      RETURN
      END SUBROUTINE MINC
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE PART
! 
! 
         USE MeshMaker_Data
         USE Grid_Gener_parameters
! 
! 
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
! 
      CHARACTER WORD*5
      CHARACTER DUAL*5
! 
      COMMON/MINCD_I/J,NVOL,MM
      COMMON/MINCD_R/VOL(200),A(200),D(200)
      COMMON/PROXI_I/L
      COMMON/PROXI_R/PAR(7)
! 
      DIMENSION WORD(16),DUAL(2)
! 
      SAVE ICALL,DUAL
      DATA DUAL/'MMVER','MMALL'/
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of PART
!
!
      ICALL=ICALL+1
      IF(ICALL.EQ.1) WRITE(*,899)
  899 FORMAT('PART              1.0   22 JANUARY   1990',6X,     &
     &       'READ SPECIFICATIONS OF MINC-PARTITIONING FROM FILE "INPUT"')
!
      READ 5020,(WORD(I),I=1,16)
 5020 FORMAT(16A5)
      IF(WORD(1).EQ.'PART ') GOTO 920
!
      PRINT 901,WORD(1)
  901 FORMAT(' HAVE READ UNKNOWN BLOCK LABEL "',A5,'" IN MINC MODULE ==> STOP EXECUTION !!!')
      STOP
!
  920 CONTINUE
!
!****READ DATA FOR MULTIPLE INTERACTING CONTINUA.***********************
!
!***** *J* IS THE NUMBER OF MULTIPLE INTERACTING CONTINUA.
!*
!***** *NVOL* (.LE.J) IS THE NUMBER OF EXPLICITLY SPECIFIED VOLUME
!*     FRACTIONS.
!*
!***** *WHE_x* SPECIFIES WHETHER EXPLICITLY PROVIDED VOLUME FRACTIONS
!      ARE GIVEN STARTING AT THE OUTSIDE (FRACTURE) OR INSIDE (MATRIX).
!*
!***** *PAR* IS AN ARRAY WITH PARAMETERS FOR SPECIFYING FRACTURE
!*     DISTRIBUTIONS.
!
      DO 902 L=1,6
      IF(WORD(2) == TYPOS(L)) GOTO 903
  902 CONTINUE
      PRINT 904,WORD(2)
  904 FORMAT(' HAVE READ UNKNOWN PROXIMITY FUNCTION IDENTIFIER  "',A5,'" ==>  STOP EXECUTION')
      STOP
!
  903 CONTINUE
!-----INDEX *L* LABELS THE TYPE OF PROXIMITY FUNCTION SELECTED.
!
!
      READ 1,J,NVOL,WHE_x,(PAR(I),I=1,7)
    1 FORMAT(2I3,A4,7E10.4)
!
!-----READ A SET OF VOLUME FRACTIONS-----
      IF(WHE_x.EQ.'OUT ') READ 2,(VOL(M),M=1,NVOL)
      IF(WHE_x.EQ.'IN  ') READ 2,(VOL(J+1-M),M=1,NVOL)
    2 FORMAT(8E10.4)
!
!----- END OF MINC-DATA ------------------------------------------------
!
      IF((L.EQ.2 .OR. L.EQ.3) .AND. PAR(2).EQ.0.) PAR(2)=PAR(1)
      IF(L.EQ.3 .AND. PAR(3).EQ.0.) PAR(3)=PAR(2)
!
!-----IDENTIFY CHOICE MADE FOR GLOBAL MATRIX-MATRIX FLOW.
!
      DO 802 MM=1,2
      IF(WORD(4) == DUAL(MM)) GOTO 803
  802 CONTINUE
      MM=0
!
  803 CONTINUE
      PRINT 801,WORD(4)
  801 FORMAT(/' CHOICE OF MATRIX-MATRIX FLOW HANDLING: "',A5,'"')
      PRINT 804
  804 FORMAT(/'                       THE OPTIONS ARE: "     "',   &
     &        ' (DEFAULT), NO GLOBAL MATRIX-MATRIX FLOW; GLOBAL FLOW ONLY THROUGH FRACTURES'/   &
     &    40X,'"MMVER", GLOBAL MATRIX-MATRIX FLOW IN VERTICAL DIRECTION ONLY'/   &
     &    40X,'"MMALL", GLOBAL MATRIX-MATRIX FLOW IN ALL DIRECTIONS'/)
      IF(MM.EQ.2.AND.J.NE.2) PRINT 806
  806 FORMAT(' !!!!! WARNING !!!!! THE "MMALL" OPTION SHOULD ONLY',   &
     &       ' BE USED WITH TWO CONTINUA, WHERE IT AMOUNTS TO A DUAL-PERMEABILITY TREATMENT')
!
  805 CONTINUE
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of PART
!
!
      RETURN
      END SUBROUTINE PART
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE GEOM
! 
! 
         USE MeshMaker_Data
         USE Grid_Gener_parameters
! 
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
! 
      DIMENSION X(200)
! 
      COMMON/MINCD_I/J,NVOL, MM
      COMMON/MINCD_R/VOL(200),A(200),D(200)
      COMMON/PROXI_I/L
      COMMON/PROXI_R/PAR(7)
! 
      SAVE ICALL,DELTA
      DATA DELTA/1.E-8/
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of GEOM
!
!
      ICALL=ICALL+1
      IF(ICALL.EQ.1) WRITE(*,899)
  899 FORMAT('GEOM              1.0    1 MAY       1991',6X,   &
     &       'CALCULATE GEOMETRY PARAMETERS OF SECONDARY (FRACTURED-POROUS) MESH')
!
      IF(NVOL.GE.J) GO TO 3
!
!-----COME HERE TO ASSIGN EQUAL VOLUMINA TO SUBDIVISIONS WHICH HAVE NOT
!     BEEN EXPLICITLY SPECIFIED-----
!
      VEX=0.0d0
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ Begin correction 
! ........ Avoid problem of stalling when WHE_x does not have the desired value
! ----------------
! 
      DO 4 M=1,NVOL
         IF(WHE_x.EQ.'OUT ') THEN
            VEX=VEX+VOL(M)
         ELSE IF(WHE_x.EQ.'IN  ') THEN
            VEX=VEX+VOL(J+1-M)
         ELSE IF(WHE_x.EQ.'IN  ') THEN
            PRINT 8100, WHE_x
            STOP
         END IF
    4 CONTINUE
 8100 FORMAT(//,20('ERROR-'),//,T33,   &
     &             '       S I M U L A T I O N   A B O R T E D',/,   &
     &         T31,'The parameter WHE_x = ',a4,   &
     &             ' is neither *OUT * nor *IN  *',   &
     &       /,T33,'               CORRECT AND TRY AGAIN',   &    
     &       //,20('ERROR-'))
! 
! ----------------
! ........ GJM - 9/19/2000
! ........ End correction 
! ----------------
! 
!     VEX IS THE TOTAL EXPLICITLY ASSIGNED VOLUME FRACTION.
!
      IF(VEX.GT.1.) GO TO 10
      IF(VEX.EQ.1.) GO TO 3
!
      VF=(1.-VEX)/FLOAT(J-NVOL)
!-----VF IS THE VOLUME FRACTION FOR PARTITIONS WHICH ARE NOT
!     EXPLICITLY ASSIGNED.
      NVOL1=NVOL+1
      DO 5 M=NVOL1,J
         IF(WHE_x.EQ.'OUT ') THEN
            VOL(M)=VF
         ELSE
            VOL(J+1-M)=VF
         END IF
    5 CONTINUE
      GO TO 3
!
   10 CONTINUE
!-----COME HERE IF EXPLICITLY ASSIGNED VOLUMINA EXCEED 100%-----
      PRINT 11,VEX
   11 FORMAT(' PROGRAM STOPS BECAUSE TOTAL VOLUME VEX = ',   &
     &       E12.5,' > 100%  ---  NEED TO CORRECT INPUT DATA')
      STOP
!
    3 CONTINUE
!
!-----NOW FIND DISTANCES FROM FRACTURES WHICH CORRESPOND TO
!     DESIRED VOLUME FRACTIONS.
!     INDEXING STARTS AT THE OUTSIDE; I.E. *1* IS THE OUTERMOST
!     VOLUME ELEMENT, AND *J* IS THE INNERMOST ONE.
!
!-----INITIALIZE TOTAL VOLUME FRACTION.
      TVOL=0.
!
!-----FIRST INTERFACE WILL BE AT FRACTURE FACE.
      X(1)=0.0d0
      D(1)=0.0d0
      A(1)=(1.0d0-VOL(1))*PROX(1.d-10)/1.d-10
!
!
!-----INITIALIZE SEARCH INTERVAL.
!
      XL = 0.0d0
      XR = VOL(2)/A(1)
!
!
      DO 30 M=2,J
!
!-----COMPUTE TOTAL FRACTION OF MATRIX VOLUME.
         TVOL = TVOL+VOL(M)/(1.-VOL(1))
         IF(M.EQ.J) TVOL = 1.-1.d-9
! 
         CALL INVER(TVOL,XMID,XL,XR)
! 
         X(M)=XMID
! 
         XMD  = XMID*DELTA
         A(M) = (1.d0-VOL(1))*(PROX(XMID+XMD)-PROX(XMID-XMD))   &
     &                       /(2.0d0*XMD)
!
         D(M) = (X(M)-X(M-1))/2.0d0
!
!-----PUT LEFT END OF NEXT ITERATION INTERVAL AT PRESENT X.
         XL = XMID
!
   30 CONTINUE
!
!
!-----COME HERE TO COMPUTE A QUASI-STEADY VALUE FOR INNERMOST
!     NODAL DISTANCE.
!
      GOTO (41,42,43,44,45,46,47,48,49,50),L
!
   41 CONTINUE
!----- ONE-D CASE.
      D(J) = (PAR(1)-2.0d0*X(J-1))/6.
      GO TO 40
!
   42 CONTINUE
!----- TWO-D CASE.
      U    = PAR(1)-2.0d0*X(J-1)
      V    = PAR(2)-2.0d0*X(J-1)
      D(J) = U*V/(4.0d0*(U+V))
      GO TO 40
!
   43 CONTINUE
!----- THRED CASE.
      U    = PAR(1)-2.0d0*X(J-1)
      V    = PAR(2)-2.0d0*X(J-1)
      W    = PAR(3)-2.0d0*X(J-1)
      D(J) = 3.0d0*U*V*W/(1.0d1*(U*V+V*W+U*W))
      GO TO 40
!
   44 CONTINUE
   45 CONTINUE
   46 CONTINUE
   47 CONTINUE
   48 CONTINUE
   49 CONTINUE
   50 CONTINUE
      D(J) = (X(J)-X(J-1))/5.0d0
!
   40 CONTINUE
!
!-----PRINT OUT GEOMETRY DATA.
!
      PRINT 27
      PRINT 23
      PRINT 24
      PRINT 25,VOL(1),D(1)
      PRINT 26,A(1),X(1)
!
   23 FORMAT('  CONTINUUM     IDENTIFIER       VOLUME      NODAL ',   &
     &       'DISTANCE      INTERFACE AREA   INTERFACE DISTANCE')
   24 FORMAT(84X,'FROM FRACTURES',/)
   25 FORMAT('  1-FRACTURES      * *    ',2(4X,E12.5))
   26 FORMAT(66X,E12.5,7X,E12.5)
   27 FORMAT(//' ==================== GEOMETRY DATA, NORMALIZED TO',   &
     &         ' A DOMAIN OF UNIT VOLUME =========================',//)
!
      DO 100 M=2,J
         PRINT 101,M,NA(M),VOL(M),D(M)
         IF(M.NE.J) PRINT 102,A(M),X(M)
  100 CONTINUE
  101 FORMAT(' ',I2,'-MATRIX',9X,'*',A1,'*',8X,E12.5,4X,E12.5)
  102 FORMAT(66X,E12.5,7X,E12.5)
!
      PRINT 103
  103 FORMAT(/,1X,99('='))
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of GEOM
!
!
      RETURN
      END SUBROUTINE GEOM
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      REAL(KIND = 8) FUNCTION PROX(X)
! 
!
!-----THE PROXIMITY FUNCTION PROX(X) REPRESENTS THE FRACTION OF
!     MATRIX VOLUME [VM=(1.-VOL(1))*V0 WITHIN A DOMAIN V0] WHICH
!     IS WITHIN A DISTANCE X FROM THE FRACTURES.
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
!
      COMMON/PROXI_I/L
      COMMON/PROXI_R/PAR(7)
      SAVE ICALL,A,B,C,D
!     NOW ASSIGN DATA FOR STANFORD LARGE RESERVOIR MODEL.
      DATA A,B,C,D/.263398,.190754,.2032,.191262/
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of PROX
!
!
      ICALL=ICALL+1
      IF(ICALL.EQ.1) WRITE(*,899)
  899 FORMAT('PROX              1.0   22 JANUARY   1990',6X,   &
     &       'CALCULATE PROXIMITY FUNCTIONS FOR DIFFERENT MATRIX BLOCK SHAPES')
!
      GOTO(1,2,3,4,4,4,1,1,1,1),L
!
    1 CONTINUE
!
!----- ONE-D CASE.
!
      PROX = 2.0d0*X/PAR(1)
      IF(X .GE. PAR(1)/2.0d0) PROX = 1.0d0
      RETURN
    2 CONTINUE
!
!----- TWO-D CASE.
!     THE MATRIX BLOCKS HAVE THICKNESS OF PAR(1) AND PAR(2),
!     RESPECTIVELY, MEASURED PERPENDICULAR TO THE FRACTURES.
!     THE PROXIMITY FUNCTION IS VALID FOR ARBITRARY ANGLE
!     BETWEEN THE FRACTURE SETS.
!
      PROX =  2.0d0*(PAR(1)+PAR(2))*X/(PAR(1)*PAR(2))   &
     &      - 4.0d0*X*X/(PAR(1)*PAR(2))
      IF(X .GE. PAR(1)/2.00d0 .OR. X .GE. PAR(2)/2.0d0) PROX = 1.0d0
      RETURN
    3 CONTINUE
!
!----- THREE DIMENSIONAL CASE.
!
      U    = 2.0d0*X/PAR(1)
      V    = 2.0d0*X/PAR(2)
      W    = 2.0d0*X/PAR(3)
      PROX = U*V*W-(U*V+U*W+V*W)+U+V+W
      IF(U .GE. 1.0d0 .OR. V .GE. 1.0d0 .OR. W .GE. 1.0d0) PROX = 1.0d0
      RETURN
    4 CONTINUE
!
!
!***** MATRIX OF STANFORD LARGE RESERVOIR MODEL *****
!
!
!     RECTANGULAR BLOCKS IN LAYERS B1,B2,M1,M2,T1.
      VR = 8.0d0*X**3-(8.0d0*B+4.0d0*A)*X**2+(4.0d0*A*B+2.0d0*B**2)*X
      IF(X .GE. B/2.0d0) VR = A*B*B
!
!     TRIANGULAR BLOCKS IN LAYERS B1,B2,M1,M2,T1.
      VT =(6.0d0+4.0d0*SQRT(2.0d0))*X**3   &
     &    -(A*(6.0d0+4.0d0*SQRT(2.0d0))/2.0d0   &
     &    +2.0d0*B*(2.0d0+SQRT(2.0d0)))*X**2   &
     &    +(A*B*(2.0d0+SQRT(2.0d0))+B*B)*X
      IF(X.GE.B/(2.0d0+SQRT(2.0d0))) VT = A*B*B/2.0d0
!
!     RECTANGULAR BLOCKS IN LAYER T2.
      VRT2 = 8.0d0*X**3-(8.0d0*D+4.0d0*C)*X**2+(4.0d0*C*D+2.0d0*D**2)*X
      IF(X .GE. D/2.0d0) VRT2 = C*D*D
!
!     TRIANGULAR BLOCKS IN LAYER T2.
      VTT2 =(6.0d0+4.0d0*SQRT(2.0d0))*X**3   &
     &      -(C*(6.0d0+4.0d0*SQRT(2.0d0))/2.0d0   &
     &      +2.0d0*D*(2.0d0+SQRT(2.0d0)))*X**2   &
     &      +(C*D*(2.0d0+SQRT(2.0d0))+D*D)*X
      IF(X.GE.D/(2.0d0+SQRT(2.0d0))) VTT2=C*D*D/2.0d0
!
      IF(L.EQ.4) GOTO 14
      IF(L.EQ.5) GOTO 15
      IF(L.EQ.6) GOTO 16
!
!***** NOW COMPUTE TOTAL MATRIX VOLUME WITHIN DISTANCE X.
   14 V = 5.0d0*(5.0d0*VR+4.0d0*VT)+5.0d0*VRT2+4.0d0*VTT2
!
!-----AVERAGE PROXIMITY FUNCTION FOR ENTIRE ROCK LOADING.
!
      VTOT =35.0d0*A*B**2+7.0d0*C*D**2
!     VOLUME FRACTION.
      PROX = V/VTOT
      RETURN
!
   15 PROX = (5.0d0*VR+4.*VT)/(7.0d0*A*B*B)
!-----PROXIMITY FUNCTION FOR FIVE BOTTOM LAYERS.
      RETURN
!
   16 PROX = (5.0d0*VRT2+4.*VTT2)/(7.0d0*C*D*D)
!-----PROXIMITY FUNCTION FOR TOP LAYER.
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of PROX
!
!
      RETURN
!
!
      END FUNCTION PROX
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE INVER(F,X,XL,XR)
! 
!
!===== THIS ROUTINE INVERTS THE PROXIMITY FUNCTION, TO GIVE A
!     DISTANCE *X* FROM FRACTURE FACES FOR A DESIRED FRACTION *F* OF
!     MATRIX VOLUME.
!
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
!
      SAVE ICALL,TOL
      DATA TOL/1.E-10/
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of INVER
!
!
      ICALL=ICALL+1
      IF(ICALL.EQ.1) WRITE(*,899)
  899 FORMAT('INVER             1.0   22 JANUARY   1990',6X,'INVERT A MONOTONIC FUNCTION THROUGH NESTED BISECTIONS')
!
!-----CHECK AND ADJUST UPPER LIMIT OF SEARCH INTERVAL.
!
   22 FR=PROX(XR)
      IF(FR.GT.F) GO TO 20
      XR=2.0d0*XR
      GO TO 22
!
!-----PERFORM ITERATIVE BISECTING, TO OBTAIN A SEQUENCE OF NESTED
!     INTERVALS CONTAINING THE DESIRED POINT, X.
!
   20 XMID=(XR+XL)/2.
      IF(XR-XL.LE.TOL*XR) GO TO 21
      FMID=PROX(XMID)
      IF(FMID.LE.F) XL=XMID
      IF(FMID.GE.F) XR=XMID
      GOTO 20
!
   21 CONTINUE
!
!-----COME HERE FOR CONVERGENCE.
!
      X = XMID
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of INVER
!
!
      RETURN
      END SUBROUTINE INVER
!
!
!
!***************************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***************************************************************************
!
!
!
      SUBROUTINE MINCME
!
! 
         USE MeshMaker_Data, MNEL => Max_NumElem
         USE Grid_Gener_parameters
! 
! 
         IMPLICIT DOUBLE PRECISION (A-H,O-Z)
!
!=====THIS ROUTINE WORKS SEQUENTIALLY THROUGH THE ELEMENTS OF THE
!     PRIMARY MESH, ASSIGNING ALL SECONDARY ELEMENTS AND INTRA-BLOCK
!     CONNECTIONS.
!
! 
      CHARACTER DENT*5
      CHARACTER EL1*1,EL2*4,EM1*1,EM2*4,EL10*1
      CHARACTER ELE*5,EC1*5,EC2*5,ELREF*5
! 
      COMMON/MINCD_I/J,NVOL, MM
      COMMON/MINCD_R/VOL(200),A(200),D(200)
      COMMON/PROXI_I/L
      COMMON/PROXI_R/PAR(7)
      COMMON/NP/NELP,NELAP,NCONP
! 
      DIMENSION DENT(16)
! 
! - GJM: Begin modification for 8-character elements
! 
      CHARACTER ELE_8*8,EREF_8*8,EC1_8*8,EC2_8*8
      CHARACTER EL1_3*3,EL2_5*5,EM1_3*3,EM2_5*5
! 
      CHARACTER MESHX*4
! 
      CHARACTER(LEN=1) :: first
      CHARACTER(LEN=3) :: activity
! 
! - GJM: End modification for 8-character elements
! 
      SAVE ICALL
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of MINCME
!
!
      ICALL=ICALL+1
      IF(ICALL.EQ.1) WRITE(*,899)
  899 FORMAT('MINCME            1.0   14 FEBRUARY  1990',6X,   &
     &       'PROCESS PRIMARY MESH FROM FILE *MESH*, AND WRITE SECONDARY MESH ON FILE *MINC*')
!
      REWIND (UNIT = MESH_Unit)
      REWIND (UNIT = MINC_Unit)
!
      PRINT 2
    2 FORMAT(/' READ PRIMARY MESH FROM FILE *MESH*')
!
!*****READ ELEMENT DATA FROM FILE *MESH*.*******************************
!
      READ(MESH_Unit,1) (DENT(I),I=1,16)
    1 FORMAT(16A5)
      IF(DENT(1).NE.'ELEME') THEN
         PRINT 4,DENT(1)
         STOP
      END IF
!
    4 FORMAT(' HAVE READ UNKNOWN BLOCK LABEL "',A5,'" ON FILE *MESH*',   &
     &       ' --- STOP EXECUTION ---')
! 
! - GJM: Begin modification for 8-character elements
! 
      IF(DENT(2)(1:4).EQ.'ext2') THEN
         IF(ElemNameLength.NE.8) THEN
            PRINT 6000, MESHX,ElemNameLength
            STOP
         END IF
         MESHX = 'ext2'
         WRITE(MINC_Unit,5006)
      ELSE
         IF(ElemNameLength.NE.5) THEN
            PRINT 6000, MESHX,ElemNameLength
            STOP
         END IF
         WRITE(MINC_Unit,6)
      END IF
!
 5006 FORMAT('ELEMEext2')
    6 FORMAT('ELEME')
!
 6000 FORMAT(//,20('ERROR-'),//,T33,   &
     &             '       S I M U L A T I O N   A B O R T E D',/,   &
     &         T2,'The mesh type inicator MESHX = ',a4,   &
     &            ' is in conflict with the # of characters in the',   &
     &            ' element name <ElemNameLength> = ',i1,   &
     &       /,T32,'               CORRECT AND TRY AGAIN',    &    
     &       //,20('ERROR-'))
! 
! - GJM: End modification for 8-character elements
! 
!
      NEL  = 0
      NELA = 0
!
      Nel_2nd  = 0
      NelA_2nd = 0
!
      num_dom = 0
!
      fractured = .FALSE.  ! ... Array operation
!
    9 CONTINUE
! 
! - GJM: Begin modification for 8-character elements
! 
      IF(MESHX.EQ.'ext2') THEN
         READ(MESH_Unit,5010) EL1_3,EL2_5,MA,VOLX,AHTX,X,Y,Z, activity
         IF(EL1_3.EQ.'   '.AND.EL2_5.EQ.'     ') GO TO 40
      ELSE
         READ(MESH_Unit,10) EL1,EL2,MA,VOLX,AHTX,X,Y,Z, activity
         IF(EL1.EQ.' '.AND.EL2.EQ.'    ') GO TO 40
      END IF
! 
 5010 FORMAT(A3,A5, 7X,I5,E10.4,E10.4,10X,3F10.3,1x,a3)
   10 FORMAT(A1,A4,10X,I5,E10.4,E10.4,10X,3E10.4,1x,a3)
! 
! - GJM: End modification for 8-character elements
! 
         NEL = NEL+1
!
!
         IF( activity(1:1) == 'I' .OR. activity(1:1) == 'V' ) THEN
            MA      = ABS(MA)
            MA1     = (MA-1)*J+1
            Nel_2nd = Nel_2nd + 1
            GO TO 8
         END IF
!
         IF( (activity(1:1) /= 'I' .OR. activity(1:1) /= 'V') .AND. VOLX <= 0.0e0 ) THEN
            MA      = ABS(MA)
            MA1     = (MA-1)*J+1
            Nel_2nd = Nel_2nd + 1
            GO TO 8
         END IF
!
         NELA =  NELA + 1
!
! ...... For domains that will remain unfractured
!
         IF( MA < 0 ) THEN
            Nel_2nd  = Nel_2nd + 1
            Nela_2nd = Nela_2nd + 1
            MA       = ABS(MA)
            MA1      = (MA-1)*J+1
            GO TO 8
         END IF
!
!-----COME HERE FOR ACTIVE ELEMENTS
!
         V   = VOL(1)*VOLX
!        MA1 = MA+1
!        MA1 = (MA-1)*3+2
         MA1 = (MA-1)*J+1
         AH  = AHTX
         IF(MM == 1) AH = VOL(1)*AHTX
! 
         Nel_2nd  = Nel_2nd + 1
         Nela_2nd = Nela_2nd + 1
! 
! - GJM: Begin modification for 8-character elements
! 
         IF(MESHX.EQ.'ext2') THEN
            i_CHAR = ICHAR(EL1_3(1:1))
            IF( i_CHAR >= 65 .OR. i_CHAR <= 90 ) THEN
               first = CHAR(i_CHAR + 32)
            ELSE IF( i_CHAR >= 97 .OR. i_CHAR <= 122 ) THEN
               first = CHAR(i_CHAR - 32)
            ELSE IF( i_CHAR == ICHAR(' ') ) THEN
               first = 'f'
            END IF
            ELE_8 = first//EL1_3(2:3)//EL2_5
            WRITE(MINC_Unit,5007) ELE_8,MA1,V,AH,X,Y,Z, activity
         ELSE
            i_CHAR = ICHAR(EL1)
            IF( i_CHAR >= 65 .OR. i_CHAR <= 90 ) THEN
               first = CHAR(i_CHAR + 32)
            ELSE IF( i_CHAR >= 97 .OR. i_CHAR <= 122 ) THEN
               first = CHAR(i_CHAR - 32)
            ELSE IF( i_CHAR == ICHAR(' ') ) THEN
               first = 'f'
            END IF
!           ELE = HB//EL2
            ELE = first//EL2
            WRITE(MINC_Unit,7) ELE,MA1,V,AH,X,Y,Z, activity
         END IF
! 
 5007 FORMAT(A8, 7X,I5,2E10.4,10X,3F10.3,1x,A3)
    7 FORMAT(A5,10X,I5,2E10.4,10X,3E10.4,1x,A3)
! 
! - GJM: End modification for 8-character elements
! 
! 
!     STORE FRACTURE ELEMENT NAME FOR LATER CROSS REFERENCING
! 
! ----------------
! ........ GJM - 2/14/2015
! ----------------
! 
         IF(MESHX == 'ext2') THEN
            name(NEL) = ELE_8
         ELSE
            name(NEL) = ELE
         END IF
! 
         fractured(NEL) = .TRUE.
! 
! ----------------
! ........ GJM - 2/14/2015
! ........ End modification 
! ----------------
! 
!-----FOR EACH PRIMARY ELEMENT, ASSIGN *J* SECONDARY ELEMENTS.
! 
! 
! - GJM: Begin modification for 8-character elements
! 
         Nel_2nd  = Nel_2nd + 1
         Nela_2nd = Nela_2nd + 1
! 
         IF(MESHX.EQ.'ext2') THEN
            DO M=2,J
               V   = VOL(M)*VOLX
!              MA2 = MA+2
!              MA2 = (MA-1)*3+3
               MA2 = (MA-1)*J + m
               AH  = 0.0e0
               IF(MM == 1) AH = VOL(M)*AHTX
               IF( J > 2 ) EL1_3(1:1) = NA(m)
               WRITE(MINC_Unit,5010) EL1_3,EL2_5,MA2,V,AH,X,Y,Z, activity
            END DO
         ELSE
            DO M=2,J
!              ELE = NA(M)//EL2
               ELE = EL1//EL2
               V   = VOL(M)*VOLX
!              MA2 = MA+2
!              MA2 = (MA-1)*3+3
               MA2 = (MA-1)*J + m
               AH  = 0.0e0
               IF(MM == 1) AH = VOL(M)*AHTX
               IF( J > 2 ) ELE = NA(m)//EL2
               WRITE(MINC_Unit,7) ELE,MA2,V,AH,X,Y,Z, activity
            END DO
         END IF
! 
! - GJM: End modification for 8-character elements
! 
!
         GO TO 9
    8    CONTINUE
!-----COME HERE FOR INACTIVE ELEMENTS. THEY WILL NOT BE SUBJECTED TO
!     THE MINC PROCEDURE, BUT WILL BE HANDLED AS A SINGLE CONTINUUM.
! 
! - GJM: Begin modification for 8-character elements
! 
         IF(MESHX.EQ.'ext2') THEN
            ELE_8 = EL1_3//EL2_5
            WRITE(MINC_Unit,5007) ELE_8,MA1,VOLX,AHTX,X,Y,Z, activity
         ELSE
            ELE = EL1//EL2
            WRITE(MINC_Unit,7) ELE,MA1,VOLX,AHTX,X,Y,Z, activity
         END IF
! 
! - GJM: End modification for 8-character elements
! 
!
         GO TO 9
!
   40 CONTINUE
 8100 FORMAT(//,20('ERROR-'),//,T33,   &
     &             '       S I M U L A T I O N   A B O R T E D',/,   &
     &         T29,'The array dimension MNEL is insufficient for the MINC needs',   &
     &       /,T33,'               CORRECT AND TRY AGAIN',   &    
     &       //,20('ERROR-'))
!
!
!
      WRITE(MINC_Unit,103)
      WRITE(MINC_Unit,104)
  103 FORMAT('     ')
  104 FORMAT('CONNE')
!
!-----NOW REDEFINE ELEMENT COUNTERS.
      NBC=0
      IF(NELA.NE.0) NBC  = NEL-NELA
      IF(NELA.EQ.0) NELA = NEL
!     PARAMETERS FOR PRIMARY MESH
      NELP  = NEL
      NELAP = NELA
!     PARAMETERS FOR SECONDARY MESH
      NELA = J*NELA
      NEL  = NELA+NBC
!
!****READ CONNECTION DATA.*********************************************
!
      N     = 0
      NCONP = 0
      READ(MESH_Unit,1) DENT(1)
      IF(DENT(1).EQ.'CONNE') GOTO 1200
      PRINT 1201,DENT(1)
 1201 FORMAT(' HAVE READ UNKNOWN BLOCK LABEL "',A5,'" ON FILE *MESH*',   &
     &       ' --- STOP EXECUTION ---')
      STOP
!
 1200 CONTINUE
! 
! - GJM: Begin modification for 8-character elements
! 
      IF(MESHX.EQ.'ext2') THEN
         READ(MESH_Unit,20) EL1_3,EL2_5,EM1_3,EM2_5,ISOT,D1,D2,AREAX,BETAX
         IF(EL1_3.EQ.'   '.AND.EL2_5.EQ.'     ') GO TO 1400
         IF(EL1_3(1:1).EQ.'+') GO TO 1400
      ELSE
         READ(MESH_Unit,20) EL1,EL2,EM1,EM2,ISOT,D1,D2,AREAX,BETAX
         IF(EL1.EQ.' '.AND.EL2.EQ.'    ') GOTO 1400
         IF(EL1.EQ.'+') GO TO 1400
      END IF
! 
 5020 FORMAT(2(A3,A5), 9X,I5,4E10.4)
   20 FORMAT(2(A1,A4),15X,I5,4E10.4)
! 
! - GJM: End modification for 8-character elements
! 
      N     = N+1
      NCONP = NCONP+1
!
!-----NOW DETERMINE ACTIVITY STATUS OF ELEMENTS AT THIS CONNECTION.
!     ASSIGN THE "WOULD-BE" FRACTURE ELEMENTS, AND SEE WHETHER THEY
!     APPEAR IN THE LIST OF ACTIVE ELEMENTS
!
      IF(MESHX.EQ.'ext2') THEN
!
         i_CHAR = ICHAR(EL1_3(1:1))
         IF( i_CHAR >= 65 .OR. i_CHAR <= 90 ) THEN
            first = CHAR(i_CHAR + 32)
         ELSE IF( i_CHAR >= 97 .OR. i_CHAR <= 122 ) THEN
            first = CHAR(i_CHAR - 32)
         ELSE IF( i_CHAR == ICHAR(' ') ) THEN
            first = 'f'
         END IF
!
         EC1_8 = first//EL1_3(2:3)//EL2_5
!
         i_CHAR = ICHAR(EM1_3(1:1))
         IF( i_CHAR >= 65 .OR. i_CHAR <= 90 ) THEN
            first = CHAR(i_CHAR + 32)
         ELSE IF( i_CHAR >= 97 .OR. i_CHAR <= 122 ) THEN
            first = CHAR(i_CHAR - 32)
         ELSE IF( i_CHAR == ICHAR(' ') ) THEN
            first = 'f'
         END IF
!
         EC2_8 = first//EM1_3(2:3)//EM2_5
!
      ELSE
!
         i_CHAR = ICHAR(EL1)
         IF( i_CHAR >= 65 .OR. i_CHAR <= 90 ) THEN
            first = CHAR(i_CHAR + 32)
         ELSE IF( i_CHAR >= 97 .OR. i_CHAR <= 122 ) THEN
            first = CHAR(i_CHAR - 32)
         ELSE IF( i_CHAR == ICHAR(' ') ) THEN
            first = 'f'
         END IF
!
         EC1 = first//EL2
!
         i_CHAR = ICHAR(EM1)
         IF( i_CHAR >= 65 .OR. i_CHAR <= 90 ) THEN
            first = CHAR(i_CHAR + 32)
         ELSE IF( i_CHAR >= 97 .OR. i_CHAR <= 122 ) THEN
            first = CHAR(i_CHAR - 32)
         ELSE IF( i_CHAR == ICHAR(' ') ) THEN
            first = 'f'
         END IF
!
         EC2 = first//EM2
!
      END IF
!
!     INITIALIZE ACTIVITY INDEX
!
      IAC = 0
!     N1  = NELAP+1
!     N2  = NELAP+1
      N1  = 0
      N2  = 0
! 
! ----------------
! ........ GJM - 2/14/2015
! ----------------
! 
      IF(MESHX == 'ext2') THEN
! 
!        DO_Loop1: DO 515 I=1,NELAP
         DO_Loop1: DO i=1,NELP
! 
            EREF_8 = name(i)
! 
            IF(EREF_8 /= EC1_8) GO TO 516
            IAC = IAC+1
            N1  = I
  516       IF(EREF_8 /= EC2_8) GO TO 517
            IAC = IAC+1
            N2  = I
  517       CONTINUE
            IF(IAC == 2) GO TO 18
! 
         END DO DO_Loop1
! 
!        IF(N1 == 0 .OR. N2 == 0) GO TO 1200
! 
         IF( .NOT. fractured(N1) ) EC1_8 = EL1_3//EL2_5
         IF( .NOT. fractured(N2) ) EC2_8 = EM1_3//EM2_5
!        IF(N1 > NELAP) EC1_8 = EL1_3//EL2_5
!        IF(N2 > NELAP) EC2_8 = EM1_3//EM2_5
! 
      ELSE
! 
!        DO_Loop2: I=1,NELAP
         DO_Loop2: DO i=1,NELP
! 
            ELREF = name(i)(1:5)
! 
            IF(ELREF /= EC1) GO TO 16
            IAC = IAC+1
            N1  = I
   16       IF(ELREF /= EC2) GO TO 17
            IAC = IAC+1
            N2  = I
   17       CONTINUE
! 
            IF(IAC == 2) GO TO 18
! 
         END DO DO_Loop2
! 
!        IF(N1 == 0 .OR. N2 == 0) GO TO 1200
! 
         IF( .NOT. fractured(N1) ) EC1 = EL1//EL2
         IF( .NOT. fractured(N2) ) EC2 = EM1//EM2
!        IF(N1 > NELAP) EC1 = EL1//EL2
!        IF(N2 > NELAP) EC2 = EM1//EM2
! 
      END IF
! 
! ----------------
! ........ GJM - 2/14/2015
! ........ End modification 
! ----------------
! 
   18 CONTINUE
!
!-----GENERATE GLOBAL FRACTURE CONNECTION DATA.
!
      IF(MESHX.EQ.'ext2') THEN
         WRITE(MINC_Unit,5105) EC1_8,EC2_8,ISOT,D1,D2,AREAX,BETAX
      ELSE
         WRITE(MINC_Unit,105) EC1,EC2,ISOT,D1,D2,AREAX,BETAX
      END IF
! 
 5105 FORMAT(2A8, 9X,I5,4E10.4)
  105 FORMAT(2A5,15X,I5,4E10.4)
!
      IF(MM == 0) GO TO 1200
!
!-----COME HERE FOR GLOBAL MATRIX-MATRIX FLOW CONNECTIONS.
!
      IF( ( ABS(BETAX) == 1.0e0 .AND. MM == 1 ) .OR. MM == 2 ) GO TO 33
      GO TO 1200
!
   33 CONTINUE
!
!     IF(N1 > NELAP .AND. N2 > NELAP) GO TO 1200
      IF( ( .NOT. fractured(N1) ) .AND. ( .NOT. fractured(N2) ) ) GO TO 1200
!
!!!!! LOOP INSERTED 11-19-84 TO ASSIGN GLOBAL CONNECTIONS BETWEEN MATRIX CONTINUA.
!
! 
! - GJM: Begin modification for 8-character elements
! 
      IF(MESHX == 'ext2') THEN
! 
         DO m = 2,J
! 
            N = N+1
! 
            IF(MM == 1) THEN
               AX = AREAX * VOL(M)
            ELSE IF(MM == 2) THEN
               AX = AREAX
            END IF
! 
            IF( J > 2 ) THEN
!              IF(N1 <= NELAP) EC1_8 = NA(m)//EL1_3(2:3)//EL2_5
!              IF(N2 <= NELAP) EC2_8 = NA(m)//EM1_3(2:3)//EM2_5
               IF( fractured(N1) ) EC1_8 = NA(m)//EL1_3(2:3)//EL2_5
               IF( fractured(N2) ) EC2_8 = NA(m)//EM1_3(2:3)//EM2_5
            ELSE
!              IF(N1 <= NELAP) EC1_8 = EL1_3//EL2_5
!              IF(N2 <= NELAP) EC2_8 = EM1_3//EM2_5
               IF( fractured(N1) ) EC1_8 = EL1_3//EL2_5
               IF( fractured(N2) ) EC2_8 = EM1_3//EM2_5
            END IF
! 
            WRITE(MINC_Unit,5105) EC1_8,EC2_8,ISOT,D1,D2,AX,BETAX
! 
         END DO
! 
      ELSE
! 
         DO m = 2,J
! 
            N = N+1
! 
            IF(MM == 1) THEN
               AX = AREAX * VOL(M)
            ELSE IF(MM == 2) THEN
               AX = AREAX
            END IF
! 
            IF( J > 2 ) THEN
!              IF(N1 <= NELAP) EC1 = NA(m)//EL2
!              IF(N2 <= NELAP) EC2 = NA(m)//EM2
               IF( fractured(N1) ) EC1 = NA(m)//EL2
               IF( fractured(N2) ) EC2 = NA(m)//EM2
            ELSE
!              IF(N1 <= NELAP) EC1 = EL1//EL2
!              IF(N2 <= NELAP) EC2 = EM1//EM2
               IF( fractured(N1) ) EC1 = EL1//EL2
               IF( fractured(N2) ) EC2 = EM1//EM2
            END IF
! 
            WRITE(MINC_Unit,105) EC1,EC2,ISOT,D1,D2,AX,BETAX
! 
         END DO
! 
      END IF
! 
! - GJM: End modification for 8-character elements
! 
!
      GO TO 1200
!
!-----END OF CONNECTION DATA.-------------------------------------------
!
 1400 CONTINUE
      PRINT 5,NELP,NELAP,NCONP
    5 FORMAT('            THE PRIMARY MESH HAS',I5,' ELEMENTS (',I5,' ACTIVE) AND ',I6,' CONNECTIONS (INTERFACES) BETWEEN THEM')
!
!-----NOW ASSIGN INTRA-BLOCK CONNECTION DATA.
!     LOOP AGAIN OVER ELEMENTS
!
      REWIND (UNIT = MESH_Unit)
      READ(MESH_Unit,1) (DENT(I),I=1,16)
      I=0
! 
! - GJM: Begin modification for 8-character elements
! 
  110 IF(MESHX.EQ.'ext2') THEN
         READ(MESH_Unit,5010) EL1_3,EL2_5,MA,VOLX,AHTX
         IF(EL1_3.EQ.'   '.AND.EL2_5.EQ.'     ') GO TO 111
      ELSE
         READ(MESH_Unit,10) EL1,EL2,MA,VOLX,AHTX
         IF(EL1.EQ.' '.AND.EL2.EQ.'    ') GO TO 111
      END IF
! 
! ... GJM: End modification for 8-character elements
! 
      I = I+1
!
!-----FOR INACTIVE ELEMENTS, DO NOT ASSIGN INTRABLOCK CONNECTIONS
!
      IF( .NOT. fractured(i) ) GO TO 110
!     IF(I > NELAP) GO TO 111
!
!-----COME HERE TO ASSIGN INTRABLOCK CONNECTIONS FOR ACTIVE ELEMENTS ONLY
!
      IF(MESHX == 'ext2') THEN
         i_CHAR = ICHAR(EL1_3(1:1))
      ELSE
         i_CHAR = ICHAR(EL1)
      END IF
! 
      IF( i_CHAR >= 65 .OR. i_CHAR <= 90 ) THEN
         first = CHAR(i_CHAR + 32)
      ELSE IF( i_CHAR >= 97 .OR. i_CHAR <= 122 ) THEN
         first = CHAR(i_CHAR - 32)
      ELSE IF( i_CHAR == ICHAR(' ') ) THEN
         first = 'f'
      END IF
! 
      BETAX = 0.0d0
! 
! - GJM: Begin modification for 8-character elements
! 
      IF(MESHX == 'ext2') THEN
! 
         DO M = 2,J
            N     = N+1
            AREAX = VOLX * A(M-1)
            IF( J > 2 ) EL1_3(1:1) = NA(m)
            WRITE(MINC_Unit,5104) first,      EL1_3(2:2),EL1_3(3:3),EL2_5,   &
     &                            EL1_3(1:1), EL1_3(2:2),EL1_3(3:3),EL2_5, D(M-1), D(M), AREAX
            IF( J > 2 ) first = NA(m)
         END DO
! 
      ELSE
! 
         DO m = 2,J
            N     = N+1
            AREAX = VOLX * A(M-1)
            IF( J > 2 ) EL1 = NA(m)
            WRITE(MINC_Unit,102) first,EL2,EL1,EL2,D(M-1),D(M),AREAX
            IF( J > 2 ) first = NA(m)
         END DO
! 
      END IF



! 
 5104 FORMAT(2(A1,A1,A1,A5),13X,'1',3E10.4)
 5102 FORMAT(2(A3,A5),13X,'1',3E10.4)
  102 FORMAT(2(A1,A4),19X,'1',3E10.4)
! 
! - GJM: Begin modification for 8-character elements
! 
      GO TO 110
!
  111 CONTINUE
      NCON=N
      PRINT 113, Nel_2nd, Nela_2nd, NCON
  113 FORMAT(/' WRITE SECONDARY MESH ON FILE *MINC*',/,   &
     &        '          THE SECONDARY MESH HAS',I5,' ELEMENTS (',I5,   &
     &        ' ACTIVE) AND ',I6,' CONNECTIONS (INTERFACES) BETWEEN THEM')
!
      WRITE(MINC_Unit,103)
      ENDFILE (UNIT = MINC_Unit)
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>   End of MINCME
!
!
      RETURN
      END SUBROUTINE MINCME
!
!
!
!***********************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***********************************************************************
!
!
!
      CHARACTER*5 FUNCTION ONOMA(NLXYZ1,NLXYZ2,NLXYZ3,i,j,k)
! 
         USE Grid_Gener_parameters
!
!*********************************************************************
!*                                                                   *
!*!           FUNCTION FOR DETERMINING THE ELEMENT NAMES             *     
!*!                      CREATED BY MESHMAKER                        *     
!*!                Version 1.00, September 18, 2000                  *     
!*                                                                   *
!*********************************************************************
!
!
!
!*********************************************************************
!*                                                                   *
!*               C O M M O N    D E C L A R A T I O N S              *     
!*                                                                   *
!*********************************************************************
!
!     COMMON/N/NA(36)
! 
      CHARACTER*1 L1,L2,L3,L4,L5
! 
      SAVE ICALL
      DATA ICALL/0/
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> Main body of ONOMA
!
!
      ICALL = ICALL+1
      IF(ICALL == 1) WRITE(*,899)
  899 FORMAT(/,'ONOMA             1.0   18 September 2000',6X,   &
     &         'FUNCTION RETURNING THE ELEMENT NAMES OF THE GRID CREATED BY MESHMAKER')
 
! 
!*********************************************************************
!*                                                                   *
!*                       DETERMINATION PROCESS                       *
!*                                                                   *
!*********************************************************************
! 
      IF(NLXYZ1.GT.1) THEN
         IF(NLXYZ1.GT.100) THEN
            nnn333 = 1+((k-1)/100)
            IF(nnn333.GT.36) THEN
               PRINT 8100
               STOP
            ELSE
               iii222 = (j-1)+(i-1)*NLXYZ2
               nnn222 = 1+MOD(iii222,36) 
               nnn111 = 1+(iii222/36)
            END IF
         ELSE
            iii333 = (j-1)+(i-1)*NLXYZ2
            nnn333 = 1+MOD(iii333,36) 
            iii222 = 1+(iii333/36)
!
            IF(iii222.EQ.36) THEN
               nnn222 = iii222
               nnn111 = nnn222/36
            ELSE IF(iii222.LT.36) THEN
               nnn222 = iii222
               nnn111 = 1+nnn222/36
            ELSE
               IF(MOD(iii222,36).NE.0) THEN
                  nnn222 = MOD(iii222,36)
                  nnn111 = 1+(iii222/36)
               ELSE
                  nnn222 = 36
                  nnn111 = iii222/36
               END IF
            END IF
!
         END IF
         mmm444 = mod(k-1,100)
         nnn444 = mmm444/10
         nnn555 = mod(mmm444,10)
      ELSE
         IF(NLXYZ2.NE.1) THEN
            IF(NLXYZ2.GT.100) THEN
               nnn333 = 1+((j-1)/100)
               IF(nnn333.GT.36) THEN
                  nnn333 = 1+MOD(nnn333,36)
                  iii333 = 1+nnn333/36
                  IF(iii333.GT.36) THEN
                     nnn222 = 1+MOD(iii333,36)
                     nnn111 = 1+nnn222/36
                  ELSE
                     nnn222 = iii333
                     nnn111 = 1
                  END IF
               ELSE
                  nnn222 = 1+MOD((i-1),36) 
                  nnn111 = 1+((i-1)/36)
               END IF
            ELSE
               nnn333 = 1+MOD((i-1),36) 
               nnn222 = 1+((i-1)/36)
               IF(nnn222.GT.36) THEN
                  nnn111 = 1+MOD(nnn222,36)
               ELSE
                  nnn111 = 1
               END IF
            END IF
            mmm444 = mod(j-1,100)
            nnn444 = mmm444/10
            nnn555 = mod(mmm444,10)
         ELSE
            IF(NLXYZ3.GT.100) THEN
               nnn333 = 1+((i-1)/100)
               IF(nnn333.GT.36) THEN
                  nnn333 = 1+MOD(nnn333,36)
                  iii333 = 1+nnn333/36
                  IF(iii333.GT.36) THEN
                     nnn222 = 1+MOD(iii333,36)
                     nnn111 = 1+nnn222/36
                  ELSE
                     nnn222 = iii333
                     nnn111 = 1
                  END IF
               ELSE
                  nnn222 = 1
                  nnn111 = 1
               END IF
            ELSE
               nnn333 = 1
               nnn222 = 1
               nnn111 = 1
            END IF
            mmm444 = mod(i-1,100)
            nnn444 = mmm444/10
            nnn555 = mod(mmm444,10)
         END IF
      END IF
! 
!***********************************************************************
!*                                                                     *
!*                     RETURNING THE onoma VALUE                       *
!*                                                                     *
!***********************************************************************
! 
      IF(nnn111+10 <= 36) THEN
         L1 = NA(nnn111+10)
      ELSE
         L1 = NA(nnn111+10-36)
      END IF
! 
      L2 = NA(nnn222) 
      L3 = NA(nnn333) 
      L5 = NA(nnn555+1) 
! 
      IF(nnn444 >= 1) THEN
         L4 = NA(nnn444+1) 
      ELSE
         L4 = '0'
      END IF
!
 1000 ONOMA = L1//L2//L3//L4//L5
!
!
!***********************************************************************
!*                                                                     *
!*                        F  O  R  M  A  T  S                          *
!*                                                                     *
!***********************************************************************
!
!
 8100 FORMAT(//,20('ERROR-'),//,T33,   &
     &             '       S I M U L A T I O N   A B O R T E D',/,   &
     &         T19,'The number of elements along the shortest ',   &
     &             'dimension of a 3-D grid exceeds 3600',   &
     &       /,T33,'               CORRECT AND TRY AGAIN',   &
     &       //,20('ERROR-'))
!
!
!  =>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=>=> End of ONOMA
!
!
 9999 RETURN
      END FUNCTION ONOMA
!
!
!
!***********************************************************************
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!***********************************************************************
!
!
!     END MODULE Grid_Generator
!
!
!
!<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
!<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
!<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
!<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


